from __future__ import annotations

import asyncio
import json
from dataclasses import dataclass
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, AsyncIterator, Optional

from dateutil import parser
from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import (
    DestinationConnectionError,
    SourceConnectionError,
    SourceConnectionNetworkError,
)
from unstructured_ingest.errors_v2 import UserAuthError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    Uploader,
    UploaderConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.processes.utils.blob_storage import (
    BlobStoreUploadStager,
    BlobStoreUploadStagerConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from office365.graph_client import GraphClient
    from office365.onedrive.driveitems.driveItem import DriveItem
    from office365.onedrive.drives.drive import Drive

CONNECTOR_TYPE = "onedrive"
MAX_BYTES_SIZE = 512_000_000


class OnedriveAccessConfig(AccessConfig):
    client_cred: str = Field(description="Microsoft App client secret")
    password: Optional[str] = Field(description="Service account password", default=None)


class OnedriveConnectionConfig(ConnectionConfig):
    client_id: str = Field(description="Microsoft app client ID")
    user_pname: str = Field(
        description="User principal name or service account, usually your Azure AD email."
    )
    tenant: str = Field(
        repr=False, description="ID or domain name associated with your Azure AD instance"
    )
    authority_url: Optional[str] = Field(
        repr=False,
        default="https://login.microsoftonline.com",
        examples=["https://login.microsoftonline.com"],
        description="Authentication token provider for Microsoft apps",
    )
    access_config: Secret[OnedriveAccessConfig]

    def get_drive(self) -> "Drive":
        client = self.get_client()
        drive = client.users[self.user_pname].drive
        return drive

    @requires_dependencies(["msal", "requests"], extras="onedrive")
    def get_token(self):
        from msal import ConfidentialClientApplication
        from requests import post

        if self.access_config.get_secret_value().password:
            url = f"https://login.microsoftonline.com/{self.tenant}/oauth2/v2.0/token"
            headers = {"Content-Type": "application/x-www-form-urlencoded"}
            data = {
                "grant_type": "password",
                "username": self.user_pname,
                "password": self.access_config.get_secret_value().password,
                "client_id": self.client_id,
                "client_secret": self.access_config.get_secret_value().client_cred,
                "scope": "https://graph.microsoft.com/.default",
            }
            response = post(url, headers=headers, data=data)
            if response.status_code == 200:
                return response.json()
            else:
                raise SourceConnectionError(
                    f"Oauth2 authentication failed with {response.status_code}: {response.text}"
                )

        else:
            try:
                app = ConfidentialClientApplication(
                    authority=f"{self.authority_url}/{self.tenant}",
                    client_id=self.client_id,
                    client_credential=self.access_config.get_secret_value().client_cred,
                )
                token = app.acquire_token_for_client(
                    scopes=["https://graph.microsoft.com/.default"]
                )
            except ValueError as exc:
                logger.error("Couldn't set up credentials.")
                raise exc

            if "error" in token:
                error_codes = token.get("error_codes", [])
                error_type = token.get("error", "")
                error_description = token.get("error_description", "")

                # 7000215: Invalid client secret provided
                # 7000218: Invalid client id provided
                # 700016: Application not found in directory
                # 90002: Tenant not found
                auth_error_codes = [7000215, 7000218, 700016, 90002]

                if any(code in error_codes for code in auth_error_codes) or error_type in [
                    "invalid_client",
                    "unauthorized_client",
                    "invalid_grant",
                ]:
                    raise UserAuthError(f"Authentication failed: {error_type}: {error_description}")
                else:
                    raise SourceConnectionNetworkError(
                        f"Failed to fetch token: {error_type}: {error_description}"
                    )
            return token

    @requires_dependencies(["office365"], extras="onedrive")
    def get_client(self) -> "GraphClient":
        from office365.graph_client import GraphClient

        client = GraphClient(self.get_token)
        return client


class OnedriveIndexerConfig(IndexerConfig):
    path: Optional[str] = Field(default="")
    recursive: bool = False


@dataclass
class OnedriveIndexer(Indexer):
    connection_config: OnedriveConnectionConfig
    index_config: OnedriveIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            token_resp: dict = self.connection_config.get_token()
            if error := token_resp.get("error"):
                raise SourceConnectionError(
                    "{} ({})".format(error, token_resp.get("error_description"))
                )
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def list_objects_sync(self, folder: DriveItem, recursive: bool) -> list["DriveItem"]:
        drive_items = folder.children.get().execute_query()
        files = [d for d in drive_items if d.is_file]
        if not recursive:
            return files

        folders = [d for d in drive_items if d.is_folder]
        for f in folders:
            files.extend(self.list_objects_sync(f, recursive))
        return files

    async def list_objects(self, folder: "DriveItem", recursive: bool) -> list["DriveItem"]:
        return await asyncio.to_thread(self.list_objects_sync, folder, recursive)

    def get_root_sync(self, client: "GraphClient") -> "DriveItem":
        root = client.users[self.connection_config.user_pname].drive.get().execute_query().root
        if fpath := self.index_config.path:
            root = root.get_by_path(fpath).get().execute_query()
            if root is None or not root.is_folder:
                raise ValueError(f"Unable to find directory, given: {fpath}")
        return root

    async def get_root(self, client: "GraphClient") -> "DriveItem":
        return await asyncio.to_thread(self.get_root_sync, client)

    def get_properties_sync(self, drive_item: "DriveItem") -> dict:
        properties = drive_item.properties
        filtered_properties = {}
        for k, v in properties.items():
            try:
                json.dumps(v)
                filtered_properties[k] = v
            except TypeError:
                pass
        return filtered_properties

    async def get_properties(self, drive_item: "DriveItem") -> dict:
        return await asyncio.to_thread(self.get_properties_sync, drive_item)

    def drive_item_to_file_data_sync(self, drive_item: "DriveItem") -> FileData:
        file_path = drive_item.parent_reference.path.split(":")[-1]
        file_path = file_path[1:] if file_path and file_path[0] == "/" else file_path
        filename = drive_item.name
        server_path = file_path + "/" + filename
        rel_path = server_path.replace(self.index_config.path, "").lstrip("/")
        date_modified_dt = (
            parser.parse(str(drive_item.last_modified_datetime))
            if drive_item.last_modified_datetime
            else None
        )
        date_created_at = (
            parser.parse(str(drive_item.created_datetime)) if drive_item.created_datetime else None
        )
        return FileData(
            identifier=drive_item.id,
            connector_type=self.connector_type,
            source_identifiers=SourceIdentifiers(
                fullpath=server_path, filename=drive_item.name, rel_path=rel_path
            ),
            metadata=FileDataSourceMetadata(
                url=drive_item.parent_reference.path + "/" + drive_item.name,
                version=drive_item.etag,
                date_modified=str(date_modified_dt.timestamp()) if date_modified_dt else None,
                date_created=str(date_created_at.timestamp()) if date_created_at else None,
                date_processed=str(time()),
                record_locator={
                    "user_pname": self.connection_config.user_pname,
                    "server_relative_path": server_path,
                },
            ),
            additional_metadata=self.get_properties_sync(drive_item=drive_item),
            display_name=server_path,
        )

    async def drive_item_to_file_data(self, drive_item: "DriveItem") -> FileData:
        # Offload the file data creation if it's not guaranteed async
        return await asyncio.to_thread(self.drive_item_to_file_data_sync, drive_item)

    def is_async(self) -> bool:
        return True

    async def run_async(self, **kwargs: Any) -> AsyncIterator[FileData]:
        token_resp = await asyncio.to_thread(self.connection_config.get_token)
        if "error" in token_resp:
            raise SourceConnectionError(
                f"[{self.connector_type}]: {token_resp['error']} "
                f"({token_resp.get('error_description')})"
            )

        client = await asyncio.to_thread(self.connection_config.get_client)
        root = await self.get_root(client=client)
        drive_items = await self.list_objects(folder=root, recursive=self.index_config.recursive)

        for drive_item in drive_items:
            file_data = await self.drive_item_to_file_data(drive_item=drive_item)
            yield file_data


class OnedriveDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class OnedriveDownloader(Downloader):
    connection_config: OnedriveConnectionConfig
    download_config: OnedriveDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    @SourceConnectionNetworkError.wrap
    def _fetch_file(self, file_data: FileData) -> DriveItem:
        if file_data.source_identifiers is None or not file_data.source_identifiers.fullpath:
            raise ValueError(
                f"file data doesn't have enough information to get "
                f"file content: {file_data.model_dump()}"
            )

        server_relative_path = file_data.source_identifiers.fullpath
        client = self.connection_config.get_client()
        root = client.users[self.connection_config.user_pname].drive.get().execute_query().root
        file = root.get_by_path(server_relative_path).get().execute_query()
        if not file:
            raise FileNotFoundError(f"file not found: {server_relative_path}")
        return file

    def get_download_path(self, file_data: FileData) -> Optional[Path]:
        rel_path = file_data.source_identifiers.relative_path
        rel_path = rel_path[1:] if rel_path.startswith("/") else rel_path
        return self.download_dir / Path(rel_path)

    @SourceConnectionError.wrap
    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        try:
            file = self._fetch_file(file_data=file_data)
            fsize = file.get_property("size", 0)
            download_path = self.get_download_path(file_data=file_data)
            download_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(f"downloading {file_data.source_identifiers.fullpath} to {download_path}")
            if fsize > MAX_BYTES_SIZE:
                logger.info(f"downloading file with size: {fsize} bytes in chunks")
                with download_path.open(mode="wb") as f:
                    file.download_session(f, chunk_size=1024 * 1024 * 100).execute_query()
            else:
                with download_path.open(mode="wb") as f:
                    file.download_session(f).execute_query()
            return self.generate_download_response(file_data=file_data, download_path=download_path)
        except Exception as e:
            logger.error(
                f"[{self.connector_type}] Exception during downloading: {e}", exc_info=True
            )
            # Re-raise to see full stack trace locally
            raise


class OnedriveUploaderConfig(UploaderConfig):
    remote_url: str = Field(
        description="URL of the destination in OneDrive, e.g., 'onedrive://Documents/Folder'"
    )
    prefix: str = "onedrive://"

    @property
    def root_folder(self) -> str:
        url = (
            self.remote_url.replace(self.prefix, "", 1)
            if self.remote_url.startswith(self.prefix)
            else self.remote_url
        )
        return url.split("/")[0]

    @property
    def url(self) -> str:
        url = (
            self.remote_url.replace(self.prefix, "", 1)
            if self.remote_url.startswith(self.prefix)
            else self.remote_url
        )
        return url


@dataclass
class OnedriveUploader(Uploader):
    connection_config: OnedriveConnectionConfig
    upload_config: OnedriveUploaderConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["office365"], extras="onedrive")
    def precheck(self) -> None:
        from office365.runtime.client_request_exception import ClientRequestException

        try:
            token_resp: dict = self.connection_config.get_token()
            if error := token_resp.get("error"):
                raise SourceConnectionError(
                    "{} ({})".format(error, token_resp.get("error_description"))
                )
            drive = self.connection_config.get_drive()
            root = drive.root
            root_folder = self.upload_config.root_folder
            folder = root.get_by_path(root_folder)
            try:
                folder.get().execute_query()
            except ClientRequestException as e:
                if not e.response.status_code == 404:
                    raise e
                folder = root.create_folder(root_folder).execute_query()
                logger.info(f"successfully created folder: {folder.name}")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["office365"], extras="onedrive")
    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        from office365.onedrive.driveitems.conflict_behavior import ConflictBehavior
        from office365.runtime.client_request_exception import ClientRequestException

        drive = self.connection_config.get_drive()

        # Use the remote_url from upload_config as the base destination folder
        base_destination_folder = self.upload_config.url
        # Use the file's relative path to maintain directory structure, if needed
        if file_data.source_identifiers and file_data.source_identifiers.relative_path:
            # Combine the base destination folder with the file's relative path
            destination_path = Path(base_destination_folder) / Path(
                f"{file_data.source_identifiers.relative_path}.json"
            )
        else:
            # If no relative path is provided, upload directly to the base destination folder
            destination_path = Path(base_destination_folder) / f"{path.name}.json"

        destination_folder = destination_path.parent
        file_name = destination_path.name

        # Convert destination folder to a string suitable for OneDrive API
        destination_folder_str = str(destination_folder).replace("\\", "/")

        # Resolve the destination folder in OneDrive, creating it if necessary
        try:
            # Attempt to get the folder
            folder = drive.root.get_by_path(destination_folder_str)
            folder.get().execute_query()
        except ClientRequestException as e:
            # Folder doesn't exist, create it recursively
            root = drive.root
            root_folder = self.upload_config.root_folder
            if not e.response.status_code == 404:
                raise e
            folder = root.create_folder(root_folder).execute_query()
            logger.info(f"successfully created folder: {folder.name}")

        # Check the size of the file
        file_size = path.stat().st_size

        if file_size < MAX_BYTES_SIZE:
            # Use simple upload for small files
            with path.open("rb") as local_file:
                content = local_file.read()
                logger.info(f"Uploading {path} to {destination_path} using simple upload")
                try:
                    uploaded_file = folder.upload(file_name, content).execute_query()
                    if not uploaded_file or uploaded_file.name != file_name:
                        raise DestinationConnectionError(f"Upload failed for file '{file_name}'")
                    # Log details about the uploaded file
                    logger.info(
                        f"Uploaded file '{uploaded_file.name}' with ID '{uploaded_file.id}'"
                    )
                except Exception as e:
                    logger.error(f"Failed to upload file '{file_name}': {e}", exc_info=True)
                    raise DestinationConnectionError(
                        f"Failed to upload file '{file_name}': {e}"
                    ) from e
        else:
            # Use resumable upload for large files
            destination_drive_item = drive.root.get_by_path(destination_folder_str)

            logger.info(
                f"Uploading {path.parent / file_name} to {destination_folder_str} using resumable upload"  # noqa: E501
            )

            try:
                uploaded_file = destination_drive_item.resumable_upload(
                    source_path=str(path)
                ).execute_query()
                # Rename the uploaded file to the original source name with a .json extension
                # Overwrite the file if it already exists
                renamed_file = uploaded_file.move(
                    name=file_name, conflict_behavior=ConflictBehavior.Replace
                ).execute_query()
                # Validate the upload
                if not renamed_file or renamed_file.name != file_name:
                    raise DestinationConnectionError(f"Upload failed for file '{file_name}'")
                # Log details about the uploaded file
                logger.info(f"Uploaded file {renamed_file.name} with ID {renamed_file.id}")
            except Exception as e:
                logger.error(f"Failed to upload file '{file_name}' using resumable upload: {e}")
                raise DestinationConnectionError(
                    f"Failed to upload file '{file_name}' using resumable upload: {e}"
                ) from e


onedrive_source_entry = SourceRegistryEntry(
    connection_config=OnedriveConnectionConfig,
    indexer_config=OnedriveIndexerConfig,
    indexer=OnedriveIndexer,
    downloader_config=OnedriveDownloaderConfig,
    downloader=OnedriveDownloader,
)

onedrive_destination_entry = DestinationRegistryEntry(
    connection_config=OnedriveConnectionConfig,
    uploader=OnedriveUploader,
    uploader_config=OnedriveUploaderConfig,
    upload_stager_config=BlobStoreUploadStagerConfig,
    upload_stager=BlobStoreUploadStager,
)



from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any, AsyncGenerator, Literal, Optional, Union

from pydantic import BaseModel, Field, HttpUrl

from unstructured_ingest.errors_v2 import ProviderError, RateLimitError, UserAuthError, UserError
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.string_and_date_utils import fix_unescaped_unicode

if TYPE_CHECKING:
    from httpx import AsyncClient, Client


class Attachment(BaseModel):
    # https://developer.zendesk.com/api-reference/ticketing/tickets/ticket-attachments/#json-format
    content_type: Optional[str] = None


class Via(BaseModel):
    # https://developer.zendesk.com/documentation/ticketing/reference-guides/via-object-reference/
    channel: Union[int, str]
    source: dict = Field(default_factory=dict)


class ZendeskComment(BaseModel):
    # https://developer.zendesk.com/api-reference/ticketing/tickets/ticket_comments/#json-format
    attachments: list[Attachment] = Field(default_factory=list)
    audit_id: Optional[int] = None
    author_id: Optional[int] = None
    body: Optional[str] = None
    created_at: Optional[datetime] = None
    html_body: Optional[str] = None
    id: Optional[int] = None
    metadata: Optional[dict] = None
    plain_body: Optional[str] = None
    public: Optional[bool] = None
    comment_type: Literal["Comment", "VoiceComment"] = Field(alias="type")
    uploads: list[str] = Field(default_factory=list)
    via: Optional[Via] = None

    def as_text(self) -> str:
        all_data = self.model_dump()
        filtered_data = {
            k: v
            for k, v in all_data.items()
            if k in ["id", "author_id", "body", "created_at"] and v is not None
        }
        return "".join(
            [f"{v}\n" for v in ["comment"] + [f"{k}: {v}" for k, v in filtered_data.items()]]
        )


class ZendeskTicket(BaseModel):
    # https://developer.zendesk.com/api-reference/ticketing/tickets/tickets/#json-format
    allow_attachments: bool = True
    allow_channelback: bool = True
    assignee_email: Optional[str] = None
    assignee_id: Optional[int] = None
    attribute_value_ids: list[int] = Field(default_factory=list)
    brand_id: Optional[int] = None
    collaborator_ids: list[int] = Field(default_factory=list)
    collaborators: list[Union[int, str, dict[str, str]]] = Field(default_factory=list)
    comment: Optional[ZendeskComment] = None
    created_at: Optional[datetime] = None
    custom_fields: list[dict[str, Any]] = Field(default_factory=list)
    custom_status_id: Optional[int] = None
    description: Optional[str] = None
    due_at: Optional[datetime] = None
    email_cc_ids: list[int] = Field(default_factory=list)
    email_ccs: list[dict[str, str]] = Field(default_factory=list)
    external_id: Optional[str] = None
    follower_ids: list[int] = Field(default_factory=list)
    followers: list[dict[str, str]] = Field(default_factory=list)
    followup_ids: list[int] = Field(default_factory=list)
    forum_topic_id: Optional[int] = None
    from_messaging_channel: bool
    generated_timestamp: Optional[datetime] = None
    group_id: Optional[int] = None
    has_incidents: bool = False
    id: Optional[int] = None
    is_public: bool = False
    macro_id: Optional[int] = None
    macro_ids: list[int] = Field(default_factory=list)
    metadata: dict[str, Any] = Field(default_factory=dict)
    organization_id: Optional[int] = None
    priority: Optional[Literal["urgent", "high", "normal", "low"]] = None
    problem_id: Optional[int] = None
    raw_subject: Optional[str] = None
    recipient: Optional[str] = None
    requester: dict[str, str] = Field(default_factory=dict)
    requester_id: int
    safe_update: Optional[bool] = None
    satisfaction_rating: Optional[Union[str, dict[str, Any]]] = None
    sharing_agreement_ids: list[int] = Field(default_factory=list)
    status: Optional[Literal["new", "open", "pending", "hold", "solved", "closed"]] = None
    subject: Optional[str] = None
    submitter_id: Optional[int] = None
    tags: list[str] = Field(default_factory=list)
    ticket_form_id: Optional[int] = None
    ticket_type: Optional[Literal["problem", "incident", "question", "task"]] = Field(
        default=None, alias="type"
    )
    updated_at: Optional[datetime] = None
    updated_stamp: Optional[str] = None
    url: Optional[HttpUrl] = None
    via: Optional[Via] = None
    via_followup_source_id: Optional[int] = None
    via_id: Optional[int] = None
    voice_comment: Optional[dict] = None

    def as_text(self) -> str:
        all_data = self.model_dump()
        filtered_data = {
            k: v
            for k, v in all_data.items()
            if k in ["id", "subject", "description", "created_at"] and v is not None
        }
        return "".join(
            [f"{v}\n" for v in ["ticket"] + [f"{k}: {v}" for k, v in filtered_data.items()]]
        )


class ZendeskArticle(BaseModel):
    # https://developer.zendesk.com/api-reference/help_center/help-center-api/articles/#json-format
    author_id: Optional[int] = None
    body: Optional[str] = None
    comments_disabled: bool = False
    content_tag_ids: list[str] = Field(default_factory=list)
    created_at: Optional[datetime] = None
    draft: bool = False
    edited_at: Optional[datetime] = None
    html_url: Optional[HttpUrl] = None
    id: int
    label_names: list[str] = Field(default_factory=list)
    locale: str
    outdated: bool = False
    outdated_locales: list[str] = Field(default_factory=list)
    permission_group_id: int
    position: Optional[int] = None
    promoted: bool = False
    section_id: Optional[int] = None
    source_locale: Optional[str] = None
    title: str
    updated_at: Optional[datetime] = None
    url: Optional[HttpUrl] = None
    user_segment_id: Optional[int] = None
    user_segment_ids: list[int] = Field(default_factory=list)
    vote_count: Optional[int] = None
    vote_sum: Optional[int] = None

    def as_html(self) -> str:
        html = self.body
        if title := self.title:
            html = f"<h1>{title}</h1>{html}"
        return fix_unescaped_unicode(f"<body class='Document' >{html}</body>")


class ZendeskArticleAttachment(BaseModel):
    # https://developer.zendesk.com/api-reference/help_center/help-center-api/article_attachments/#json-format
    article_id: Optional[int] = None
    content_type: Optional[str] = None
    content_url: Optional[HttpUrl] = None
    created_at: Optional[datetime] = None
    guide_media_id: Optional[str] = None
    id: Optional[int] = None
    inline: bool = False
    locale: Optional[str] = None
    size: Optional[int] = None
    updated_at: Optional[datetime] = None
    url: Optional[HttpUrl] = None


@dataclass
class ZendeskClient:
    token: str
    subdomain: str
    email: str
    max_page_size: int = 100
    _async_client: "AsyncClient" = field(init=False, default=None)
    _client: "Client" = field(init=False, default=None)
    _base_url: str = field(init=False, default=None)

    async def __aenter__(self) -> "ZendeskClient":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._async_client.aclose()

    @requires_dependencies(["httpx"], extras="zendesk")
    def __post_init__(self):
        import httpx

        auth = f"{self.email}/token", self.token
        self._client = httpx.Client(auth=auth)
        self._async_client = httpx.AsyncClient(auth=auth)
        self._base_url = f"https://{self.subdomain}.zendesk.com/api/v2"

        # Run check
        try:
            url_to_check = f"{self._base_url}/groups.json"
            resp = self._client.head(url_to_check)
            resp.raise_for_status()
        except Exception as e:
            raise self.wrap_error(e=e)

    @requires_dependencies(["httpx"], extras="zendesk")
    def wrap_error(self, e: Exception) -> Exception:
        import httpx

        if not isinstance(e, httpx.HTTPStatusError):
            logger.error(f"unhandled exception from Zendesk client: {e}", exc_info=True)
            return e
        url = e.request.url
        response_code = e.response.status_code
        if response_code == 401:
            logger.error(
                f"Failed to connect via auth,"
                f"{url} using zendesk response, status code {response_code}"
            )
            return UserAuthError(e)
        if response_code == 429:
            logger.error(
                f"Failed to connect via rate limits"
                f"{url} using zendesk response, status code {response_code}"
            )
            return RateLimitError(e)
        if 400 <= response_code < 500:
            logger.error(
                f"Failed to connect to {url} using zendesk response, status code {response_code}"
            )
            return UserError(e)
        if response_code > 500:
            logger.error(
                f"Failed to connect to {url} using zendesk response, status code {response_code}"
            )
            return ProviderError(e)
        logger.error(f"unhandled http status error from Zendesk client: {e}", exc_info=True)
        return e

    async def fetch_content(self, url: str, content_key: str) -> AsyncGenerator[dict, None]:
        url = f"{url}?page[size]={self.max_page_size}"
        while True:
            try:
                response = await self._async_client.get(url)
                response.raise_for_status()
            except Exception as e:
                raise self.wrap_error(e=e)

            data = response.json()
            for content in data[content_key]:
                yield content

            has_more = data.get("meta", {}).get("has_more", False)
            if not has_more:
                break

            url = data["links"]["next"]

    async def get_articles(self) -> AsyncGenerator[ZendeskArticle, None]:
        """
        Retrieves article content from Zendesk asynchronously.
        """
        article_url = f"https://{self.subdomain}.zendesk.com/api/v2/help_center/articles.json"

        try:
            async for article_dict in self.fetch_content(url=article_url, content_key="articles"):
                yield ZendeskArticle.model_validate(article_dict)
        except Exception as e:
            raise self.wrap_error(e=e)

    async def get_comments(self, ticket_id: int) -> AsyncGenerator[ZendeskComment, None]:
        comments_url = f"https://{self.subdomain}.zendesk.com/api/v2/tickets/{ticket_id}/comments"

        try:
            async for comment_dict in self.fetch_content(url=comments_url, content_key="comments"):
                yield ZendeskComment.model_validate(comment_dict)
        except Exception as e:
            raise self.wrap_error(e=e)

    async def get_tickets(self) -> AsyncGenerator[ZendeskTicket, None]:
        tickets_url = f"https://{self.subdomain}.zendesk.com/api/v2/tickets"

        try:
            async for ticket_dict in self.fetch_content(url=tickets_url, content_key="tickets"):
                yield ZendeskTicket.model_validate(ticket_dict)
        except Exception as e:
            raise self.wrap_error(e=e)

    async def get_article_attachments(
        self, article_id: int
    ) -> AsyncGenerator[ZendeskArticleAttachment, None]:
        """
        Handles article attachments such as images and stores them as UTF-8 encoded bytes.
        """
        article_attachment_url = (
            f"https://{self.subdomain}.zendesk.com/api/v2/help_center/"
            f"articles/{article_id}/attachments"
        )

        try:
            async for attachment_dict in self.fetch_content(
                url=article_attachment_url, content_key="article_attachments"
            ):
                yield ZendeskArticleAttachment.model_validate(attachment_dict)
        except Exception as e:
            raise self.wrap_error(e=e)



"""
Filesystem utilities for concurrent operations.

This module provides race-condition-safe filesystem operations that are needed
when multiple processes operate on the same directory structures simultaneously.
"""

from pathlib import Path


def mkdir_concurrent_safe(path: Path) -> None:
    """
    Create directory safely in concurrent environments, handling race conditions.

    This addresses the issue where Path.mkdir(parents=True, exist_ok=True) can still
    raise FileExistsError when multiple processes attempt to create overlapping
    directory structures simultaneously. In this codebase, this occurs when multiple
    files are being downloaded in parallel and archive extraction is happening in parallel.

    Related: https://github.com/python/cpython/pull/112966/files
    Python core team used the same approach to fix zipfile race conditions.
    """
    try:
        path.mkdir(parents=True, exist_ok=True)
    except FileExistsError:
        if not (path.exists() and path.is_dir()):
            raise



# https://developers.notion.com/reference/property-object#rich-text
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import Div, HtmlTag, Span

from unstructured_ingest.processes.connectors.notion.interfaces import DBCellBase, DBPropertyBase
from unstructured_ingest.processes.connectors.notion.types.rich_text import (
    RichText as RichTextType,
)


@dataclass
class RichText(DBPropertyBase):
    id: str
    name: str
    type: str = "rich_text"
    description: Optional[str] = None
    rich_text: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class RichTextCell(DBCellBase):
    id: str
    rich_text: List[RichTextType]
    name: Optional[str] = None
    type: str = "rich_text"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            rich_text=[RichTextType.from_dict(rt) for rt in data.pop("rich_text", [])],
            **data,
        )

    def get_html(self) -> Optional[HtmlTag]:
        if not self.rich_text:
            return None
        spans = [Span([], rt.get_html()) for rt in self.rich_text]
        return Div([], spans)



import json
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, ContextManager, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import (
    DestinationConnectionError,
    SourceConnectionError,
    SourceConnectionNetworkError,
)
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    Uploader,
    UploaderConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.data_prep import batch_generator
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from confluent_kafka import Consumer, Producer


class KafkaAccessConfig(AccessConfig, ABC):
    pass


class KafkaConnectionConfig(ConnectionConfig, ABC):
    access_config: Secret[KafkaAccessConfig]
    bootstrap_server: str
    port: int
    group_id: str = Field(
        description="A consumer group is a way to allow a pool of consumers "
        "to divide the consumption of data over topics and partitions.",
        default="default_group_id",
    )

    @abstractmethod
    def get_consumer_configuration(self) -> dict:
        pass

    @abstractmethod
    def get_producer_configuration(self) -> dict:
        pass

    @contextmanager
    @requires_dependencies(["confluent_kafka"], extras="kafka")
    def get_consumer(self) -> ContextManager["Consumer"]:
        from confluent_kafka import Consumer

        consumer = Consumer(self.get_consumer_configuration())
        try:
            logger.debug("kafka consumer connected")
            yield consumer
        finally:
            consumer.close()

    @requires_dependencies(["confluent_kafka"], extras="kafka")
    def get_producer(self) -> "Producer":
        from confluent_kafka import Producer

        producer = Producer(self.get_producer_configuration())
        return producer


class KafkaIndexerConfig(IndexerConfig):
    topic: str = Field(description="which topic to consume from")
    num_messages_to_consume: Optional[int] = 100
    timeout: Optional[float] = Field(default=3.0, description="polling timeout", ge=3.0)

    def update_consumer(self, consumer: "Consumer") -> None:
        consumer.subscribe([self.topic])


@dataclass
class KafkaIndexer(Indexer, ABC):
    connection_config: KafkaConnectionConfig
    index_config: KafkaIndexerConfig

    @contextmanager
    def get_consumer(self) -> ContextManager["Consumer"]:
        with self.connection_config.get_consumer() as consumer:
            self.index_config.update_consumer(consumer=consumer)
            yield consumer

    @requires_dependencies(["confluent_kafka"], extras="kafka")
    def generate_messages(self) -> Generator[Any, None, None]:
        from confluent_kafka import KafkaError, KafkaException

        messages_consumed = 0
        max_empty_polls = 10
        empty_polls = 0
        num_messages_to_consume = self.index_config.num_messages_to_consume
        with self.get_consumer() as consumer:
            while messages_consumed < num_messages_to_consume and empty_polls < max_empty_polls:
                msg = consumer.poll(timeout=self.index_config.timeout)
                if msg is None:
                    logger.debug("No Kafka messages found")
                    empty_polls += 1
                    continue
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        logger.info(
                            "Reached end of partition for topic %s [%d] at offset %d"
                            % (msg.topic(), msg.partition(), msg.offset())
                        )
                        break
                    else:
                        raise KafkaException(msg.error())
                try:
                    empty_polls = 0
                    messages_consumed += 1
                    yield msg
                finally:
                    consumer.commit(asynchronous=False)

    def generate_file_data(self, msg) -> FileData:
        msg_content = msg.value().decode("utf8")
        identifier = f"{msg.topic()}_{msg.partition()}_{msg.offset()}"
        additional_metadata = {
            "topic": msg.topic(),
            "partition": msg.partition(),
            "offset": msg.offset(),
            "content": msg_content,
        }
        filename = f"{identifier}.txt"
        return FileData(
            identifier=identifier,
            connector_type=self.connector_type,
            source_identifiers=SourceIdentifiers(
                filename=filename,
                fullpath=filename,
            ),
            metadata=FileDataSourceMetadata(
                date_processed=str(time()),
            ),
            additional_metadata=additional_metadata,
            display_name=filename,
        )

    def run(self) -> Generator[FileData, None, None]:
        for message in self.generate_messages():
            yield self.generate_file_data(message)

    async def run_async(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        raise NotImplementedError()

    def precheck(self):
        try:
            with self.get_consumer() as consumer:
                # timeout needs at least 3 secs, more info:
                # https://forum.confluent.io/t/kafkacat-connect-failure-to-confcloud-ssl/2513
                cluster_meta = consumer.list_topics(timeout=5)
                current_topics = [
                    topic for topic in cluster_meta.topics if topic != "__consumer_offsets"
                ]
                if self.index_config.topic not in current_topics:
                    raise SourceConnectionError(
                        "expected topic '{}' not detected in cluster: '{}'".format(
                            self.index_config.topic, ", ".join(current_topics)
                        )
                    )
                logger.info(f"successfully checked available topics: {current_topics}")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")


class KafkaDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class KafkaDownloader(Downloader, ABC):
    connection_config: KafkaConnectionConfig
    download_config: KafkaDownloaderConfig = field(default_factory=KafkaDownloaderConfig)
    version: Optional[str] = None
    source_url: Optional[str] = None

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        source_identifiers = file_data.source_identifiers
        if source_identifiers is None:
            raise ValueError("FileData is missing source_identifiers")

        # Build the download path using source_identifiers
        download_path = Path(self.download_dir) / source_identifiers.relative_path
        download_path.parent.mkdir(parents=True, exist_ok=True)

        try:
            content = file_data.additional_metadata["content"]
            with open(download_path, "w") as file:
                file.write(content)
        except Exception as e:
            logger.error(f"Failed to download file {file_data.identifier}: {e}")
            raise SourceConnectionNetworkError(f"failed to download file {file_data.identifier}")

        return self.generate_download_response(file_data=file_data, download_path=download_path)


class KafkaUploaderConfig(UploaderConfig):
    batch_size: int = Field(default=100, description="Batch size")
    topic: str = Field(description="which topic to write to")
    timeout: Optional[float] = Field(
        default=10.0, description="Timeout in seconds to flush batch of messages"
    )


@dataclass
class KafkaUploader(Uploader, ABC):
    connection_config: KafkaConnectionConfig
    upload_config: KafkaUploaderConfig

    def precheck(self):
        try:
            with self.connection_config.get_consumer() as consumer:
                cluster_meta = consumer.list_topics(timeout=self.upload_config.timeout)
                current_topics = [
                    topic for topic in cluster_meta.topics if topic != "__consumer_offsets"
                ]
                logger.info(f"successfully checked available topics: {current_topics}")
                if self.upload_config.topic not in current_topics:
                    raise DestinationConnectionError(
                        "expected topic '{}' not detected in cluster: '{}'".format(
                            self.upload_config.topic, ", ".join(current_topics)
                        )
                    )

        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def produce_batch(self, elements: list[dict]) -> None:
        from confluent_kafka.error import KafkaException

        producer = self.connection_config.get_producer()
        failed_producer = False

        def acked(err, msg):
            nonlocal failed_producer
            if err is not None:
                failed_producer = True
                logger.error("Failed to deliver kafka message: %s: %s" % (str(msg), str(err)))

        for element in elements:
            producer.produce(
                topic=self.upload_config.topic,
                value=json.dumps(element),
                callback=acked,
            )

        while producer_len := len(producer):
            logger.debug(f"another iteration of kafka producer flush. Queue length: {producer_len}")
            producer.flush(timeout=self.upload_config.timeout)
        if failed_producer:
            raise KafkaException("failed to produce all messages in batch")

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        for element_batch in batch_generator(data, batch_size=self.upload_config.batch_size):
            self.produce_batch(elements=element_batch)



import inspect
from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass, field, fields
from typing import Any, Optional, Type, TypeVar

import click
from pydantic import BaseModel

from unstructured_ingest.cli.base.importer import import_from_string
from unstructured_ingest.cli.utils.click import extract_config
from unstructured_ingest.cli.utils.model_conversion import options_from_base_model, post_check
from unstructured_ingest.interfaces import ProcessorConfig
from unstructured_ingest.logger import logger
from unstructured_ingest.pipeline.pipeline import Pipeline
from unstructured_ingest.processes.chunker import Chunker, ChunkerConfig
from unstructured_ingest.processes.connector_registry import (
    DownloaderT,
    IndexerT,
    RegistryEntry,
    UploaderT,
    UploadStager,
    UploadStagerConfig,
    UploadStagerT,
    destination_registry,
    source_registry,
)
from unstructured_ingest.processes.connectors.local import LocalUploader, LocalUploaderConfig
from unstructured_ingest.processes.embedder import Embedder, EmbedderConfig
from unstructured_ingest.processes.filter import Filterer, FiltererConfig
from unstructured_ingest.processes.partitioner import Partitioner, PartitionerConfig

CommandT = TypeVar("CommandT", bound=click.Command)


@dataclass
class BaseCmd(ABC):
    cmd_name: str
    registry_entry: RegistryEntry
    default_configs: list[Type[BaseModel]] = field(default_factory=list)

    @abstractmethod
    def get_registry_options(self):
        pass

    def get_default_options(self) -> list[click.Option]:
        options = []
        for extra in self.default_configs:
            options.extend(options_from_base_model(model=extra))
        return options

    @classmethod
    def consolidate_options(cls, options: list[click.Option]) -> list[click.Option]:
        option_names = [option.name for option in options]
        duplicate_names = [name for name, count in Counter(option_names).items() if count > 1]
        if not duplicate_names:
            return options
        consolidated_options = []
        current_names = []
        for option in options:
            if option.name not in current_names:
                current_names.append(option.name)
                consolidated_options.append(option)
                continue
            existing_option = next(o for o in consolidated_options if o.name == option.name)
            if existing_option.__dict__ == option.__dict__:
                continue
            option_diff = cls.get_options_diff(o1=option, o2=existing_option)
            raise ValueError(
                "Conflicting duplicate {} option defined: {}".format(
                    option.name, " | ".join([f"{d[0]}: {d[1]}" for d in option_diff])
                )
            )
        return consolidated_options

    @staticmethod
    def get_options_diff(o1: click.Option, o2: click.Option):
        o1_dict = o1.__dict__
        o2_dict = o2.__dict__
        for d in [o1_dict, o2_dict]:
            d["opts"] = ",".join(d["opts"])
            d["secondary_opts"] = ",".join(d["secondary_opts"])
        option_diff = set(o1_dict.items()) ^ set(o2_dict.items())
        return option_diff

    @property
    def cmd_name_key(self):
        return self.cmd_name.replace("-", "_")

    @property
    def cli_cmd_name(self):
        return self.cmd_name.replace("_", "-")

    @abstractmethod
    def cmd(self, ctx: click.Context, **options) -> None:
        pass

    def add_options(self, cmd: CommandT) -> CommandT:
        options = self.get_registry_options()
        options.extend(self.get_default_options())
        post_check(options=options, name=cmd.name)
        cmd.params.extend(options)
        return cmd

    def get_pipeline(
        self,
        src: str,
        source_options: dict[str, Any],
        dest: Optional[str] = None,
        destination_options: Optional[dict[str, Any]] = None,
    ) -> Pipeline:
        logger.debug(
            f"creating pipeline from cli using source {src} with options: {source_options}"
        )
        pipeline_kwargs: dict[str, Any] = {
            "context": self.get_processor_config(options=source_options),
            "downloader": self.get_downloader(src=src, options=source_options),
            "indexer": self.get_indexer(src=src, options=source_options),
            "partitioner": self.get_partitioner(options=source_options),
        }
        if chunker := self.get_chunker(options=source_options):
            pipeline_kwargs["chunker"] = chunker
        if filterer := self.get_filterer(options=source_options):
            pipeline_kwargs["filterer"] = filterer
        if embedder := self.get_embedder(options=source_options):
            pipeline_kwargs["embedder"] = embedder
        if dest:
            logger.debug(
                f"setting destination on pipeline {dest} with options: {destination_options}"
            )
            if uploader_stager := self.get_upload_stager(dest=dest, options=destination_options):
                pipeline_kwargs["stager"] = uploader_stager
            pipeline_kwargs["uploader"] = self.get_uploader(dest=dest, options=destination_options)
        else:
            # Default to local uploader
            # TODO remove after v1 no longer supported
            destination_options = destination_options or {}
            if "output_dir" not in destination_options:
                destination_options["output_dir"] = source_options["output_dir"]
            pipeline_kwargs["uploader"] = self.get_default_uploader(options=destination_options)
        return Pipeline(**pipeline_kwargs)

    @staticmethod
    def get_default_uploader(options: dict[str, Any]) -> UploaderT:
        uploader_config = extract_config(flat_data=options, config=LocalUploaderConfig)
        return LocalUploader(upload_config=uploader_config)

    @staticmethod
    def get_chunker(options: dict[str, Any]) -> Optional[Chunker]:
        chunker_config = extract_config(flat_data=options, config=ChunkerConfig)
        if not chunker_config.chunking_strategy:
            return None
        return Chunker(config=chunker_config)

    @staticmethod
    def get_filterer(options: dict[str, Any]) -> Optional[Filterer]:
        filterer_configs = extract_config(flat_data=options, config=FiltererConfig)
        if not filterer_configs.model_dump():
            return None
        return Filterer(config=filterer_configs)

    @staticmethod
    def get_embedder(options: dict[str, Any]) -> Optional[Embedder]:
        embedder_config = extract_config(flat_data=options, config=EmbedderConfig)
        if not embedder_config.embedding_provider:
            return None
        return Embedder(config=embedder_config)

    @staticmethod
    def get_partitioner(options: dict[str, Any]) -> Partitioner:
        partitioner_config = extract_config(flat_data=options, config=PartitionerConfig)
        return Partitioner(config=partitioner_config)

    @staticmethod
    def get_processor_config(options: dict[str, Any]) -> ProcessorConfig:
        return extract_config(flat_data=options, config=ProcessorConfig)

    @staticmethod
    def get_indexer(src: str, options: dict[str, Any]) -> IndexerT:
        source_entry = source_registry[src]
        indexer_kwargs: dict[str, Any] = {}
        if indexer_config_cls := source_entry.indexer_config:
            indexer_kwargs["index_config"] = extract_config(
                flat_data=options, config=indexer_config_cls
            )
        if connection_config_cls := source_entry.connection_config:
            indexer_kwargs["connection_config"] = extract_config(
                flat_data=options, config=connection_config_cls
            )
        indexer_cls = source_entry.indexer
        return indexer_cls(**indexer_kwargs)

    @staticmethod
    def get_downloader(src: str, options: dict[str, Any]) -> DownloaderT:
        source_entry = source_registry[src]
        downloader_kwargs: dict[str, Any] = {}
        if downloader_config_cls := source_entry.downloader_config:
            downloader_kwargs["download_config"] = extract_config(
                flat_data=options, config=downloader_config_cls
            )
        if connection_config_cls := source_entry.connection_config:
            downloader_kwargs["connection_config"] = extract_config(
                flat_data=options, config=connection_config_cls
            )
        downloader_cls = source_entry.downloader
        return downloader_cls(**downloader_kwargs)

    @staticmethod
    def get_custom_stager(
        stager_reference: str, stager_config_kwargs: Optional[dict] = None
    ) -> Optional[UploadStagerT]:
        uploader_cls = import_from_string(stager_reference)
        if not inspect.isclass(uploader_cls):
            raise ValueError(
                f"custom stager must be a reference to a python class, got: {type(uploader_cls)}"
            )
        if not issubclass(uploader_cls, UploadStager):
            raise ValueError(
                "custom stager must be an implementation of the UploadStager interface"
            )
        fields_dict = {f.name: f.type for f in fields(uploader_cls)}
        upload_stager_config_cls = fields_dict["upload_stager_config"]
        if not inspect.isclass(upload_stager_config_cls):
            raise ValueError(
                f"custom stager config must be a class, got: {type(upload_stager_config_cls)}"
            )
        if not issubclass(upload_stager_config_cls, UploadStagerConfig):
            raise ValueError(
                "custom stager config must be an implementation "
                "of the UploadStagerUploadStagerConfig interface"
            )
        upload_stager_kwargs: dict[str, Any] = {}
        if stager_config_kwargs:
            upload_stager_kwargs["upload_stager_config"] = upload_stager_config_cls(
                **stager_config_kwargs
            )
        return uploader_cls(**upload_stager_kwargs)

    @staticmethod
    def get_upload_stager(dest: str, options: dict[str, Any]) -> Optional[UploadStagerT]:
        if custom_stager := options.get("custom_stager"):
            return BaseCmd.get_custom_stager(
                stager_reference=custom_stager,
                stager_config_kwargs=options.get("custom_stager_config_kwargs"),
            )
        dest_entry = destination_registry[dest]
        upload_stager_kwargs: dict[str, Any] = {}
        if upload_stager_config_cls := dest_entry.upload_stager_config:
            upload_stager_kwargs["upload_stager_config"] = extract_config(
                flat_data=options, config=upload_stager_config_cls
            )
        if upload_stager_cls := dest_entry.upload_stager:
            return upload_stager_cls(**upload_stager_kwargs)
        return None

    @staticmethod
    def get_uploader(dest, options: dict[str, Any]) -> UploaderT:
        dest_entry = destination_registry[dest]
        uploader_kwargs: dict[str, Any] = {}
        if uploader_config_cls := dest_entry.uploader_config:
            uploader_kwargs["upload_config"] = extract_config(
                flat_data=options, config=uploader_config_cls
            )
        if connection_config_cls := dest_entry.connection_config:
            uploader_kwargs["connection_config"] = extract_config(
                flat_data=options, config=connection_config_cls
            )
        uploader_cls = dest_entry.uploader
        return uploader_cls(**uploader_kwargs)



# https://developers.notion.com/reference/block#numbered-list-item
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import HtmlTag, Li

from unstructured_ingest.processes.connectors.notion.interfaces import BlockBase
from unstructured_ingest.processes.connectors.notion.types.rich_text import RichText


@dataclass
class NumberedListItem(BlockBase):
    color: str
    children: List[dict] = field(default_factory=list)
    rich_text: List[RichText] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        rich_text = data.pop("rich_text", [])
        numbered_list = cls(**data)
        numbered_list.rich_text = [RichText.from_dict(rt) for rt in rich_text]
        return numbered_list

    def get_html(self) -> Optional[HtmlTag]:
        return Li([], [rt.get_html() for rt in self.rich_text])



import json
import os
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.interfaces import (
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
)
from unstructured_ingest.processes.connectors.databricks.volumes import DatabricksPathMixin
from unstructured_ingest.processes.connectors.sql.databricks_delta_tables import (
    DatabricksDeltaTablesConnectionConfig,
    DatabricksDeltaTablesUploadStagerConfig,
)
from unstructured_ingest.utils.constants import RECORD_ID_LABEL
from unstructured_ingest.utils.data_prep import get_enhanced_element_id, get_json_data, write_data

CONNECTOR_TYPE = "databricks_volume_delta_tables"

if TYPE_CHECKING:
    pass


class DatabricksVolumeDeltaTableUploaderConfig(UploaderConfig, DatabricksPathMixin):
    database: str = Field(description="Database name", default="default")
    table_name: Optional[str] = Field(description="Table name", default=None)


class DatabricksVolumeDeltaTableStagerConfig(UploadStagerConfig):
    pass


@dataclass
class DatabricksVolumeDeltaTableStager(UploadStager):
    upload_stager_config: DatabricksVolumeDeltaTableStagerConfig = field(
        default_factory=DatabricksVolumeDeltaTableStagerConfig
    )

    def run(
        self,
        elements_filepath: Path,
        output_dir: Path,
        output_filename: str,
        file_data: FileData,
        **kwargs: Any,
    ) -> Path:
        # To avoid new line issues when migrating from volumes into delta tables, omit indenting
        # and always write it as a json file
        output_dir.mkdir(exist_ok=True, parents=True)
        output_path = output_dir / output_filename
        final_output_path = output_path.with_suffix(".json")
        data = get_json_data(path=elements_filepath)
        for element in data:
            element["id"] = get_enhanced_element_id(element_dict=element, file_data=file_data)
            element[RECORD_ID_LABEL] = file_data.identifier
            element["metadata"] = json.dumps(element.get("metadata", {}))
        write_data(path=final_output_path, data=data, indent=None)
        return final_output_path


@dataclass
class DatabricksVolumeDeltaTableUploader(Uploader):
    connection_config: DatabricksDeltaTablesConnectionConfig
    upload_config: DatabricksVolumeDeltaTableUploaderConfig
    connector_type: str = CONNECTOR_TYPE
    _columns: Optional[dict[str, str]] = None

    def init(self, **kwargs: Any) -> None:
        self.create_destination(**kwargs)

    def create_destination(
        self, destination_name: str = "unstructuredautocreated", **kwargs: Any
    ) -> bool:
        table_name = self.upload_config.table_name or destination_name
        self.upload_config.table_name = table_name
        connectors_dir = Path(__file__).parents[1]
        collection_config_file = connectors_dir / "assets" / "databricks_delta_table_schema.sql"
        with self.get_cursor() as cursor:
            cursor.execute("SHOW TABLES")
            table_names = [r[1] for r in cursor.fetchall()]
            if table_name in table_names:
                return False
            with collection_config_file.open() as schema_file:
                data_lines = schema_file.readlines()
            data_lines[0] = data_lines[0].replace("elements", table_name)
            destination_schema = "".join([line.strip() for line in data_lines])
            logger.info(f"creating table {table_name} for user")
            cursor.execute(destination_schema)
            return True

    def precheck(self) -> None:
        with self.connection_config.get_cursor() as cursor:
            cursor.execute("SHOW CATALOGS")
            catalogs = [r[0] for r in cursor.fetchall()]
            if self.upload_config.catalog not in catalogs:
                raise ValueError(
                    "Catalog {} not found in {}".format(
                        self.upload_config.catalog, ", ".join(catalogs)
                    )
                )
            cursor.execute(f"USE CATALOG '{self.upload_config.catalog}'")
            cursor.execute("SHOW DATABASES")
            databases = [r[0] for r in cursor.fetchall()]
            if self.upload_config.database not in databases:
                raise ValueError(
                    "Database {} not found in {}".format(
                        self.upload_config.database, ", ".join(databases)
                    )
                )

    def get_output_path(self, file_data: FileData, suffix: str = ".json") -> str:
        filename = Path(file_data.source_identifiers.filename)
        adjusted_filename = filename if filename.suffix == suffix else f"{filename}{suffix}"
        return os.path.join(self.upload_config.path, f"{adjusted_filename}")

    @contextmanager
    def get_cursor(self, **connect_kwargs) -> Generator[Any, None, None]:
        with self.connection_config.get_cursor(**connect_kwargs) as cursor:
            logger.debug(f"executing: USE CATALOG: '{self.upload_config.catalog}'")
            cursor.execute(f"USE CATALOG '{self.upload_config.catalog}'")
            logger.debug(f"executing: USE DATABASE: {self.upload_config.database}")
            cursor.execute(f"USE DATABASE {self.upload_config.database}")
            yield cursor

    def get_table_columns(self) -> dict[str, str]:
        if self._columns is None:
            with self.get_cursor() as cursor:
                cursor.execute(f"SELECT * from `{self.upload_config.table_name}` LIMIT 1")
                self._columns = {desc[0]: desc[1] for desc in cursor.description}
        return self._columns

    def can_delete(self) -> bool:
        existing_columns = self.get_table_columns()
        return RECORD_ID_LABEL in existing_columns

    def delete_previous_content(self, file_data: FileData) -> None:
        logger.debug(
            f"deleting any content with metadata "
            f"{RECORD_ID_LABEL}={file_data.identifier} "
            f"from delta table: {self.upload_config.table_name}"
        )
        with self.get_cursor() as cursor:
            cursor.execute(
                f"DELETE FROM `{self.upload_config.table_name}` WHERE {RECORD_ID_LABEL} = '{file_data.identifier}'"  # noqa: E501
            )
            results = cursor.fetchall()
            deleted_rows = results[0][0]
            logger.debug(f"deleted {deleted_rows} rows from table {self.upload_config.table_name}")

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        if self.can_delete():
            self.delete_previous_content(file_data=file_data)
        with self.get_cursor(staging_allowed_local_path=path.parent.as_posix()) as cursor:
            catalog_path = self.get_output_path(file_data=file_data)
            logger.debug(f"uploading {path.as_posix()} to {catalog_path}")
            cursor.execute(f"PUT '{path.as_posix()}' INTO '{catalog_path}' OVERWRITE")
            logger.debug(
                f"migrating content from {catalog_path} to table {self.upload_config.table_name}"
            )
            data = get_json_data(path=path)
            columns = data[0].keys()
            select_columns = ["PARSE_JSON(metadata)" if c == "metadata" else c for c in columns]
            column_str = ", ".join(columns)
            select_column_str = ", ".join(select_columns)
            sql_statment = f"INSERT INTO `{self.upload_config.table_name}` ({column_str}) SELECT {select_column_str} FROM json.`{catalog_path}`"  # noqa: E501
            cursor.execute(sql_statment)


databricks_volumes_delta_tables_destination_entry = DestinationRegistryEntry(
    connection_config=DatabricksDeltaTablesConnectionConfig,
    uploader=DatabricksVolumeDeltaTableUploader,
    uploader_config=DatabricksVolumeDeltaTableUploaderConfig,
    upload_stager=DatabricksVolumeDeltaTableStager,
    upload_stager_config=DatabricksDeltaTablesUploadStagerConfig,
)



import hashlib
import time
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, List

from pydantic import BaseModel, Field, Secret

from unstructured_ingest.data_types.file_data import (
    BatchFileData,
    BatchItem,
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import (
    DestinationConnectionError,
    SourceConnectionError,
    SourceConnectionNetworkError,
)
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
    download_responses,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.utils.data_prep import batch_generator, flatten_dict
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from couchbase.cluster import Cluster
    from couchbase.collection import Collection

CONNECTOR_TYPE = "couchbase"
SERVER_API_VERSION = "1"


class CouchbaseAdditionalMetadata(BaseModel):
    bucket: str


class CouchbaseBatchFileData(BatchFileData):
    additional_metadata: CouchbaseAdditionalMetadata


class CouchbaseAccessConfig(AccessConfig):
    password: str = Field(description="The password for the Couchbase server")


class CouchbaseConnectionConfig(ConnectionConfig):
    username: str = Field(description="The username for the Couchbase server")
    bucket: str = Field(description="The bucket to connect to on the Couchbase server")
    connection_string: str = Field(
        default="couchbase://localhost", description="The connection string of the Couchbase server"
    )
    scope: str = Field(
        default="_default", description="The scope to connect to on the Couchbase server"
    )
    collection: str = Field(
        default="_default", description="The collection to connect to on the Couchbase server"
    )
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    access_config: Secret[CouchbaseAccessConfig]

    @requires_dependencies(["couchbase"], extras="couchbase")
    @contextmanager
    def get_client(self) -> Generator["Cluster", None, None]:
        from couchbase.auth import PasswordAuthenticator
        from couchbase.cluster import Cluster
        from couchbase.options import ClusterOptions

        auth = PasswordAuthenticator(self.username, self.access_config.get_secret_value().password)
        options = ClusterOptions(auth)
        options.apply_profile("wan_development")
        cluster = None
        try:
            cluster = Cluster(self.connection_string, options)
            cluster.wait_until_ready(timedelta(seconds=5))
            yield cluster
        finally:
            if cluster:
                cluster.close()


class CouchbaseUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class CouchbaseUploadStager(UploadStager):
    upload_stager_config: CouchbaseUploadStagerConfig = field(
        default_factory=lambda: CouchbaseUploadStagerConfig()
    )

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        data = element_dict.copy()
        return {
            data["element_id"]: {
                "embedding": data.get("embeddings", None),
                "text": data.get("text", None),
                "metadata": data.get("metadata", None),
                "type": data.get("type", None),
            }
        }


class CouchbaseUploaderConfig(UploaderConfig):
    batch_size: int = Field(default=50, description="Number of documents to upload per batch")


@dataclass
class CouchbaseUploader(Uploader):
    connection_config: CouchbaseConnectionConfig
    upload_config: CouchbaseUploaderConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            self.connection_config.get_client()
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        logger.info(
            f"writing {len(data)} objects to destination "
            f"bucket, {self.connection_config.bucket} "
            f"at {self.connection_config.connection_string}",
        )
        with self.connection_config.get_client() as client:
            bucket = client.bucket(self.connection_config.bucket)
            scope = bucket.scope(self.connection_config.scope)
            collection = scope.collection(self.connection_config.collection)

            for chunk in batch_generator(data, self.upload_config.batch_size):
                collection.upsert_multi(
                    {doc_id: doc for doc in chunk for doc_id, doc in doc.items()}
                )


class CouchbaseIndexerConfig(IndexerConfig):
    batch_size: int = Field(default=50, description="Number of documents to index per batch")


@dataclass
class CouchbaseIndexer(Indexer):
    connection_config: CouchbaseConnectionConfig
    index_config: CouchbaseIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            self.connection_config.get_client()
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["couchbase"], extras="couchbase")
    def _get_doc_ids(self) -> List[str]:
        query = (
            f"SELECT META(d).id "
            f"FROM `{self.connection_config.bucket}`."
            f"`{self.connection_config.scope}`."
            f"`{self.connection_config.collection}` as d"
        )

        max_attempts = 5
        attempts = 0
        while attempts < max_attempts:
            try:
                with self.connection_config.get_client() as client:
                    result = client.query(query)
                    document_ids = [row["id"] for row in result]
                    return document_ids
            except Exception as e:
                attempts += 1
                time.sleep(3)
                if attempts == max_attempts:
                    raise SourceConnectionError(f"failed to get document ids: {e}")

    def run(self, **kwargs: Any) -> Generator[CouchbaseBatchFileData, None, None]:
        ids = self._get_doc_ids()
        for batch in batch_generator(ids, self.index_config.batch_size):
            # Make sure the hash is always a positive number to create identified
            yield CouchbaseBatchFileData(
                connector_type=CONNECTOR_TYPE,
                metadata=FileDataSourceMetadata(
                    url=f"{self.connection_config.connection_string}/"
                    f"{self.connection_config.bucket}",
                    date_processed=str(time.time()),
                ),
                additional_metadata=CouchbaseAdditionalMetadata(
                    bucket=self.connection_config.bucket
                ),
                batch_items=[BatchItem(identifier=b) for b in batch],
            )


class CouchbaseDownloaderConfig(DownloaderConfig):
    collection_id: str = Field(
        default="id", description="The unique key of the id field in the collection"
    )
    fields: list[str] = field(default_factory=list)


@dataclass
class CouchbaseDownloader(Downloader):
    connection_config: CouchbaseConnectionConfig
    download_config: CouchbaseDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    def is_async(self) -> bool:
        return False

    def get_identifier(self, bucket: str, record_id: str) -> str:
        f = f"{bucket}-{record_id}"
        if self.download_config.fields:
            f = "{}-{}".format(
                f,
                hashlib.sha256(",".join(self.download_config.fields).encode()).hexdigest()[:8],
            )
        return f

    def map_cb_results(self, cb_results: dict) -> str:
        doc_body = cb_results
        flattened_dict = flatten_dict(dictionary=doc_body)
        str_values = [str(value) for value in flattened_dict.values()]
        concatenated_values = "\n".join(str_values)
        return concatenated_values

    def generate_download_response(
        self, result: dict, bucket: str, file_data: CouchbaseBatchFileData
    ) -> DownloadResponse:
        record_id = result[self.download_config.collection_id]
        filename_id = self.get_identifier(bucket=bucket, record_id=record_id)
        filename = f"{filename_id}.txt"
        download_path = self.download_dir / Path(filename)
        logger.debug(
            f"Downloading results from bucket {bucket} and id {record_id} to {download_path}"
        )
        download_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            with open(download_path, "w", encoding="utf8") as f:
                f.write(self.map_cb_results(cb_results=result))
        except Exception as e:
            logger.error(
                f"failed to download from bucket {bucket} "
                f"and id {record_id} to {download_path}: {e}",
                exc_info=True,
            )
            raise SourceConnectionNetworkError(f"failed to download file {file_data.identifier}")
        file_data.source_identifiers = SourceIdentifiers(filename=filename, fullpath=filename)
        cast_file_data = FileData.cast(file_data=file_data)
        cast_file_data.identifier = filename_id
        cast_file_data.metadata.date_processed = str(time.time())
        cast_file_data.metadata.record_locator = {
            "connection_string": self.connection_config.connection_string,
            "bucket": bucket,
            "scope": self.connection_config.scope,
            "collection": self.connection_config.collection,
            "document_id": record_id,
        }
        return super().generate_download_response(
            file_data=cast_file_data,
            download_path=download_path,
        )

    def run(self, file_data: FileData, **kwargs: Any) -> download_responses:
        couchbase_file_data = CouchbaseBatchFileData.cast(file_data=file_data)
        bucket_name: str = couchbase_file_data.additional_metadata.bucket
        ids: list[str] = [item.identifier for item in couchbase_file_data.batch_items]

        with self.connection_config.get_client() as client:
            bucket = client.bucket(bucket_name)
            scope = bucket.scope(self.connection_config.scope)
            collection = scope.collection(self.connection_config.collection)

            download_resp = self.process_all_doc_ids(ids, collection, bucket_name, file_data)
            return list(download_resp)

    def process_doc_id(
        self,
        doc_id: str,
        collection: "Collection",
        bucket_name: str,
        file_data: CouchbaseBatchFileData,
    ):
        result = collection.get(doc_id)
        return self.generate_download_response(
            result=result.content_as[dict], bucket=bucket_name, file_data=file_data
        )

    def process_all_doc_ids(
        self,
        ids: list[str],
        collection: "Collection",
        bucket_name: str,
        file_data: CouchbaseBatchFileData,
    ):
        for doc_id in ids:
            yield self.process_doc_id(doc_id, collection, bucket_name, file_data)

    async def run_async(self, file_data: FileData, **kwargs: Any) -> download_responses:
        raise NotImplementedError()


couchbase_destination_entry = DestinationRegistryEntry(
    connection_config=CouchbaseConnectionConfig,
    uploader=CouchbaseUploader,
    uploader_config=CouchbaseUploaderConfig,
    upload_stager=CouchbaseUploadStager,
    upload_stager_config=CouchbaseUploadStagerConfig,
)

couchbase_source_entry = SourceRegistryEntry(
    connection_config=CouchbaseConnectionConfig,
    indexer=CouchbaseIndexer,
    indexer_config=CouchbaseIndexerConfig,
    downloader=CouchbaseDownloader,
    downloader_config=CouchbaseDownloaderConfig,
)



import json
import re
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import date, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional

from dateutil import parser
from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import DestinationConnectionError, WriteError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
    VectorDBUploader,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.constants import RECORD_ID_LABEL
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from weaviate.classes.init import Timeout
    from weaviate.client import WeaviateClient
    from weaviate.collections.batch.client import BatchClient

CONNECTOR_TYPE = "weaviate"


class WeaviateAccessConfig(AccessConfig, ABC):
    pass


class WeaviateConnectionConfig(ConnectionConfig, ABC):
    init_timeout: int = Field(default=2, ge=0, description="Timeout for initialization checks")
    insert_timeout: int = Field(default=90, ge=0, description="Timeout for insert operations")
    query_timeout: int = Field(default=30, ge=0, description="Timeout for query operations")
    access_config: Secret[WeaviateAccessConfig] = Field(
        default=WeaviateAccessConfig(), validate_default=True
    )

    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_timeout(self) -> "Timeout":
        from weaviate.classes.init import Timeout

        return Timeout(init=self.init_timeout, query=self.query_timeout, insert=self.insert_timeout)

    @abstractmethod
    @contextmanager
    def get_client(self) -> Generator["WeaviateClient", None, None]:
        pass


class WeaviateUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class WeaviateUploadStager(UploadStager):
    upload_stager_config: WeaviateUploadStagerConfig = field(
        default_factory=lambda: WeaviateUploadStagerConfig()
    )

    @staticmethod
    def parse_date_string(date_string: str) -> date:
        try:
            timestamp = float(date_string)
            return datetime.fromtimestamp(timestamp)
        except Exception as e:
            logger.debug(f"date {date_string} string not a timestamp: {e}")
        return parser.parse(date_string)

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        """
        Updates the element dictionary to conform to the Weaviate schema
        """
        data = element_dict.copy()
        working_data = data.copy()
        # Dict as string formatting
        if (
            record_locator := working_data.get("metadata", {})
            .get("data_source", {})
            .get("record_locator")
        ):
            # Explicit casting otherwise fails schema type checking
            working_data["metadata"]["data_source"]["record_locator"] = str(
                json.dumps(record_locator)
            )

        # Array of items as string formatting
        if points := working_data.get("metadata", {}).get("coordinates", {}).get("points"):
            working_data["metadata"]["coordinates"]["points"] = str(json.dumps(points))

        if links := working_data.get("metadata", {}).get("links", {}):
            working_data["metadata"]["links"] = str(json.dumps(links))

        if permissions_data := (
            working_data.get("metadata", {}).get("data_source", {}).get("permissions_data")
        ):
            working_data["metadata"]["data_source"]["permissions_data"] = json.dumps(
                permissions_data
            )

        # Datetime formatting
        if (
            date_created := working_data.get("metadata", {})
            .get("data_source", {})
            .get("date_created")
        ):
            working_data["metadata"]["data_source"]["date_created"] = self.parse_date_string(
                date_created
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )

        if (
            date_modified := working_data.get("metadata", {})
            .get("data_source", {})
            .get("date_modified")
        ):
            working_data["metadata"]["data_source"]["date_modified"] = self.parse_date_string(
                date_modified
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )

        if (
            date_processed := working_data.get("metadata", {})
            .get("data_source", {})
            .get("date_processed")
        ):
            working_data["metadata"]["data_source"]["date_processed"] = self.parse_date_string(
                date_processed
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )

        if last_modified := working_data.get("metadata", {}).get("last_modified"):
            working_data["metadata"]["last_modified"] = self.parse_date_string(
                last_modified
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )

        # String casting
        if version := working_data.get("metadata", {}).get("data_source", {}).get("version"):
            working_data["metadata"]["data_source"]["version"] = str(version)

        if page_number := working_data.get("metadata", {}).get("page_number"):
            working_data["metadata"]["page_number"] = str(page_number)

        if regex_metadata := working_data.get("metadata", {}).get("regex_metadata"):
            working_data["metadata"]["regex_metadata"] = str(json.dumps(regex_metadata))

        working_data[RECORD_ID_LABEL] = file_data.identifier
        return working_data


class WeaviateUploaderConfig(UploaderConfig):
    collection: Optional[str] = Field(
        description="The name of the collection this object belongs to", default=None
    )
    batch_size: Optional[int] = Field(default=None, description="Number of records per batch")
    requests_per_minute: Optional[int] = Field(default=None, description="Rate limit for upload")
    dynamic_batch: bool = Field(default=True, description="Whether to use dynamic batch")
    record_id_key: str = Field(
        default=RECORD_ID_LABEL,
        description="searchable key to find entries for the same record on previous runs",
    )

    def model_post_init(self, __context: Any) -> None:
        batch_types = {
            "fixed_size": self.batch_size is not None,
            "rate_limited": self.requests_per_minute is not None,
            "dynamic": self.dynamic_batch,
        }

        enabled_batch_modes = [batch_key for batch_key, flag in batch_types.items() if flag]
        if not enabled_batch_modes:
            raise ValueError("No batch mode enabled")
        if len(enabled_batch_modes) > 1:
            raise ValueError(
                "Multiple batch modes enabled, only one mode can be used: {}".format(
                    ", ".join(enabled_batch_modes)
                )
            )
        logger.info(f"Uploader config instantiated with {enabled_batch_modes[0]} batch mode")

    @contextmanager
    def get_batch_client(self, client: "WeaviateClient") -> Generator["BatchClient", None, None]:
        if self.dynamic_batch:
            with client.batch.dynamic() as batch_client:
                yield batch_client
        elif self.batch_size:
            with client.batch.fixed_size(batch_size=self.batch_size) as batch_client:
                yield batch_client
        elif self.requests_per_minute:
            with client.batch.rate_limit(
                requests_per_minute=self.requests_per_minute
            ) as batch_client:
                yield batch_client
        else:
            raise ValueError("No batch mode enabled")


@dataclass
class WeaviateUploader(VectorDBUploader, ABC):
    upload_config: WeaviateUploaderConfig
    connection_config: WeaviateConnectionConfig

    def _collection_exists(self, collection_name: Optional[str] = None):
        collection_name = collection_name or self.upload_config.collection
        with self.connection_config.get_client() as weaviate_client:
            return weaviate_client.collections.exists(name=collection_name)

    def precheck(self) -> None:
        try:
            self.connection_config.get_client()
            # only if collection name populated should we check that it exists
            if self.upload_config.collection and not self._collection_exists():
                raise DestinationConnectionError(
                    f"collection '{self.upload_config.collection}' does not exist"
                )
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def init(self, **kwargs: Any) -> None:
        self.create_destination(**kwargs)

    def format_destination_name(self, destination_name: str) -> str:
        """
        Weaviate Collection naming conventions:
            1. must begin with an uppercase letter
            2. must be alphanumeric and underscores only
        """

        # Check if the first character is an uppercase letter
        if not re.match(r"^[a-zA-Z]", destination_name):
            raise ValueError("Collection name must start with an uppercase letter")
        # Replace all non-alphanumeric characters with underscores
        formatted = re.sub(r"[^a-zA-Z0-9]", "_", destination_name)
        # Make the first character uppercase and leave the rest as is
        if len(formatted) == 1:
            formatted = formatted.capitalize()
        else:
            formatted = formatted[0].capitalize() + formatted[1:]
        if formatted != destination_name:
            logger.warning(
                f"Given Collection name '{destination_name}' doesn't follow naming conventions. "
                f"Renaming to '{formatted}'"
            )
        return formatted

    def create_destination(
        self,
        destination_name: str = "Unstructuredautocreated",
        vector_length: Optional[int] = None,
        **kwargs: Any,
    ) -> bool:
        collection_name = self.upload_config.collection or destination_name
        collection_name = self.format_destination_name(collection_name)
        self.upload_config.collection = collection_name

        if not self._collection_exists():
            connectors_dir = Path(__file__).parents[1]
            collection_config_file = connectors_dir / "assets" / "weaviate_collection_config.json"
            with collection_config_file.open() as f:
                collection_config = json.load(f)
            collection_config["class"] = collection_name

            logger.info(f"Creating weaviate collection '{collection_name}' with default configs")
            with self.connection_config.get_client() as weaviate_client:
                weaviate_client.collections.create_from_dict(config=collection_config)
                return True
        logger.debug(f"Collection with name '{collection_name}' already exists, skipping creation")
        return False

    def check_for_errors(self, client: "WeaviateClient") -> None:
        failed_uploads = client.batch.failed_objects
        if failed_uploads:
            for failure in failed_uploads:
                logger.error(
                    f"Failed to upload object with id {failure.original_uuid}: {failure.message}"
                )
            raise WriteError("Failed to upload to weaviate")

    @requires_dependencies(["weaviate"], extras="weaviate")
    def delete_by_record_id(self, client: "WeaviateClient", file_data: FileData) -> None:
        from weaviate.classes.query import Filter

        record_id = file_data.identifier
        collection = client.collections.get(self.upload_config.collection)
        delete_filter = Filter.by_property(name=self.upload_config.record_id_key).equal(
            val=record_id
        )
        # There is a configurable maximum limit (QUERY_MAXIMUM_RESULTS) on the number of
        # objects that can be deleted in a single query (default 10,000). To delete
        # more objects than the limit, re-run the query until nothing is deleted.
        while True:
            resp = collection.data.delete_many(where=delete_filter)
            if resp.failed:
                raise WriteError(
                    f"failed to delete records in collection "
                    f"{self.upload_config.collection} with record "
                    f"id property {record_id}"
                )
            if not resp.failed and not resp.successful:
                break

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        logger.info(
            f"writing {len(data)} objects to destination "
            f"class {self.connection_config.access_config} "
        )
        if not self.upload_config.collection:
            raise ValueError("No collection specified")

        with self.connection_config.get_client() as weaviate_client:
            self.delete_by_record_id(client=weaviate_client, file_data=file_data)
            with self.upload_config.get_batch_client(client=weaviate_client) as batch_client:
                for e in data:
                    vector = e.pop("embeddings", None)
                    batch_client.add_object(
                        collection=self.upload_config.collection,
                        properties=e,
                        vector=vector,
                    )
            self.check_for_errors(client=weaviate_client)



from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import BaseModel, Field, Secret

from unstructured_ingest.__version__ import __version__ as unstructured_version
from unstructured_ingest.data_types.file_data import (
    BatchFileData,
    BatchItem,
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    Uploader,
    UploaderConfig,
    download_responses,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.utils.constants import RECORD_ID_LABEL
from unstructured_ingest.utils.data_prep import batch_generator, flatten_dict
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from pymongo import MongoClient
    from pymongo.collection import Collection

CONNECTOR_TYPE = "mongodb"
SERVER_API_VERSION = "1"


class MongoDBAdditionalMetadata(BaseModel):
    database: str
    collection: str


class MongoDBBatchFileData(BatchFileData):
    additional_metadata: MongoDBAdditionalMetadata


class MongoDBAccessConfig(AccessConfig):
    uri: Optional[str] = Field(default=None, description="URI to user when connecting")


class MongoDBConnectionConfig(ConnectionConfig):
    access_config: Secret[MongoDBAccessConfig] = Field(
        default=MongoDBAccessConfig(), validate_default=True
    )
    host: Optional[str] = Field(
        default=None,
        description="hostname or IP address or Unix domain socket path of a single mongod or "
        "mongos instance to connect to, or a list of hostnames",
    )
    port: int = Field(default=27017)
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    @contextmanager
    @requires_dependencies(["pymongo"], extras="mongodb")
    def get_client(self) -> Generator["MongoClient", None, None]:
        from pymongo import MongoClient
        from pymongo.driver_info import DriverInfo
        from pymongo.server_api import ServerApi

        access_config = self.access_config.get_secret_value()
        if uri := access_config.uri:
            client_kwargs = {
                "host": uri,
                "server_api": ServerApi(version=SERVER_API_VERSION),
                "driver": DriverInfo(name="unstructured", version=unstructured_version),
            }
        else:
            client_kwargs = {
                "host": self.host,
                "port": self.port,
                "server_api": ServerApi(version=SERVER_API_VERSION),
            }
        with MongoClient(**client_kwargs) as client:
            # UnsupportedDigestmodError means that SCRAM-SHA-1 is disabled
            # It uses md5 which is unavailable on FIPS images
            try:
                from hashlib import UnsupportedDigestmodError  # type: ignore[attr-defined]
            except ImportError:
                from _hashlib import UnsupportedDigestmodError  # type: ignore[attr-defined]

            # Check if the authentication mechanism is supported
            try:
                client.admin.command("ping")
            except UnsupportedDigestmodError as e:
                raise ConnectionError(
                    "Authentication using SCRAM-SHA-1 is disabled. "
                    "Use SCRAM-SHA-256 instead. "
                    "See: https://www.mongodb.com/docs/manual/core/security-scram/"
                ) from e
            yield client


class MongoDBIndexerConfig(IndexerConfig):
    batch_size: int = Field(default=100, description="Number of records per batch")
    database: Optional[str] = Field(default=None, description="database name to connect to")
    collection: Optional[str] = Field(default=None, description="collection name to connect to")


class MongoDBDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class MongoDBIndexer(Indexer):
    connection_config: MongoDBConnectionConfig
    index_config: MongoDBIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        """Validates the connection to the MongoDB server."""
        try:
            with self.connection_config.get_client() as client:
                client.admin.command("ping")
                database_names = client.list_database_names()
                database_name = self.index_config.database
                if database_name not in database_names:
                    raise SourceConnectionError(
                        "database {} does not exist: {}".format(
                            database_name, ", ".join(database_names)
                        )
                    )
                database = client[database_name]
                collection_names = database.list_collection_names()
                collection_name = self.index_config.collection
                if collection_name not in collection_names:
                    raise SourceConnectionError(
                        "collection {} does not exist: {}".format(
                            collection_name, ", ".join(collection_names)
                        )
                    )
        except Exception as e:
            logger.error(f"Failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to validate connection: {e}")

    def run(self, **kwargs: Any) -> Generator[BatchFileData, None, None]:
        """Generates FileData objects for each document in the MongoDB collection."""
        with self.connection_config.get_client() as client:
            database = client[self.index_config.database]
            collection = database[self.index_config.collection]

            # Get list of document IDs
            ids = collection.distinct("_id")

        ids = sorted(ids)
        batch_size = self.index_config.batch_size

        for id_batch in batch_generator(ids, batch_size=batch_size):
            # Make sure the hash is always a positive number to create identifier
            display_name = (
                f"{self.index_config.database}.{self.index_config.collection}, "
                f"batch {id_batch[0]}-{id_batch[-1]}"
            )
            metadata = FileDataSourceMetadata(
                date_processed=str(time()),
                record_locator={
                    "database": self.index_config.database,
                    "collection": self.index_config.collection,
                },
            )

            file_data = MongoDBBatchFileData(
                connector_type=self.connector_type,
                metadata=metadata,
                batch_items=[BatchItem(identifier=str(doc_id)) for doc_id in id_batch],
                additional_metadata=MongoDBAdditionalMetadata(
                    collection=self.index_config.collection, database=self.index_config.database
                ),
                display_name=display_name,
            )
            yield file_data


@dataclass
class MongoDBDownloader(Downloader):
    download_config: MongoDBDownloaderConfig
    connection_config: MongoDBConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def generate_download_response(
        self, doc: dict, file_data: MongoDBBatchFileData
    ) -> DownloadResponse:
        from bson.objectid import ObjectId

        doc_id = doc["_id"]
        doc.pop("_id", None)

        # Extract date_created from the document or ObjectId
        date_created = None
        if "date_created" in doc:
            # If the document has a 'date_created' field, use it
            date_created = doc["date_created"]
            if isinstance(date_created, datetime):
                date_created = date_created.isoformat()
            else:
                # Convert to ISO format if it's a string
                date_created = str(date_created)
        elif isinstance(doc_id, ObjectId):
            # Use the ObjectId's generation time
            date_created = doc_id.generation_time.isoformat()

        flattened_dict = flatten_dict(dictionary=doc)
        concatenated_values = "\n".join(str(value) for value in flattened_dict.values())

        # Create a FileData object for each document with source_identifiers
        filename = f"{doc_id}.txt"
        file_data.source_identifiers = SourceIdentifiers(
            filename=filename,
            fullpath=filename,
        )
        cast_file_data = FileData.cast(file_data=file_data)
        cast_file_data.identifier = str(doc_id)

        # Determine the download path
        download_path = self.get_download_path(file_data=cast_file_data)
        if download_path is None:
            raise ValueError("Download path could not be determined")

        download_path.parent.mkdir(parents=True, exist_ok=True)

        # Write the concatenated values to the file
        with open(download_path, "w", encoding="utf8") as f:
            f.write(concatenated_values)

        # Update metadata
        cast_file_data.metadata.record_locator["document_id"] = str(doc_id)
        cast_file_data.metadata.date_created = date_created

        return super().generate_download_response(
            file_data=cast_file_data, download_path=download_path
        )

    @SourceConnectionError.wrap
    @requires_dependencies(["bson"], extras="mongodb")
    def run(self, file_data: FileData, **kwargs: Any) -> download_responses:
        """Fetches the document from MongoDB and writes it to a file."""
        from bson.errors import InvalidId
        from bson.objectid import ObjectId

        mongo_file_data = MongoDBBatchFileData.cast(file_data=file_data)

        with self.connection_config.get_client() as client:
            database = client[mongo_file_data.additional_metadata.database]
            collection = database[mongo_file_data.additional_metadata.collection]

            ids = [item.identifier for item in mongo_file_data.batch_items]

            object_ids = []
            for doc_id in ids:
                try:
                    object_ids.append(ObjectId(doc_id))
                except InvalidId as e:
                    error_message = f"Invalid ObjectId for doc_id '{doc_id}': {str(e)}"
                    logger.error(error_message)
                    raise ValueError(error_message) from e

            try:
                docs = list(collection.find({"_id": {"$in": object_ids}}))
            except Exception as e:
                logger.error(f"Failed to fetch documents: {e}", exc_info=True)
                raise e

        download_responses = []
        for doc in docs:
            download_responses.append(
                self.generate_download_response(doc=doc, file_data=mongo_file_data)
            )

        return download_responses


class MongoDBUploaderConfig(UploaderConfig):
    batch_size: int = Field(default=100, description="Number of records per batch")
    database: Optional[str] = Field(default=None, description="database name to connect to")
    collection: Optional[str] = Field(default=None, description="collection name to connect to")
    record_id_key: str = Field(
        default=RECORD_ID_LABEL,
        description="searchable key to find entries for the same record on previous runs",
    )


@dataclass
class MongoDBUploader(Uploader):
    upload_config: MongoDBUploaderConfig
    connection_config: MongoDBConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            with self.connection_config.get_client() as client:
                client.admin.command("ping")
                database_names = client.list_database_names()
                database_name = self.upload_config.database
                if database_name not in database_names:
                    raise DestinationConnectionError(
                        "database {} does not exist: {}".format(
                            database_name, ", ".join(database_names)
                        )
                    )
                database = client[database_name]
                collection_names = database.list_collection_names()
                collection_name = self.upload_config.collection
                if collection_name not in collection_names:
                    raise DestinationConnectionError(
                        "collection {} does not exist: {}".format(
                            collection_name, ", ".join(collection_names)
                        )
                    )
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def can_delete(self, collection: "Collection") -> bool:
        indexed_keys = []
        for index in collection.list_indexes():
            key_bson = index["key"]
            indexed_keys.extend(key_bson.keys())
        return self.upload_config.record_id_key in indexed_keys

    def delete_by_record_id(self, collection: "Collection", file_data: FileData) -> None:
        logger.debug(
            f"deleting any content with metadata "
            f"{self.upload_config.record_id_key}={file_data.identifier} "
            f"from collection: {collection.name}"
        )
        query = {self.upload_config.record_id_key: file_data.identifier}
        delete_results = collection.delete_many(filter=query)
        logger.info(
            f"deleted {delete_results.deleted_count} records from collection {collection.name}"
        )

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        logger.info(
            f"writing {len(data)} objects to destination "
            f"db, {self.upload_config.database}, "
            f"collection {self.upload_config.collection} "
            f"at {self.connection_config.host}",
        )
        # This would typically live in the stager but since no other manipulation
        # is done, setting the record id field in the uploader
        for element in data:
            element[self.upload_config.record_id_key] = file_data.identifier
        with self.connection_config.get_client() as client:
            db = client[self.upload_config.database]
            collection = db[self.upload_config.collection]
            if self.can_delete(collection=collection):
                self.delete_by_record_id(file_data=file_data, collection=collection)
            else:
                logger.warning("criteria for deleting previous content not met, skipping")
            for chunk in batch_generator(data, self.upload_config.batch_size):
                collection.insert_many(chunk)


mongodb_destination_entry = DestinationRegistryEntry(
    connection_config=MongoDBConnectionConfig,
    uploader=MongoDBUploader,
    uploader_config=MongoDBUploaderConfig,
)

mongodb_source_entry = SourceRegistryEntry(
    connection_config=MongoDBConnectionConfig,
    indexer_config=MongoDBIndexerConfig,
    indexer=MongoDBIndexer,
    downloader_config=MongoDBDownloaderConfig,
    downloader=MongoDBDownloader,
)



# https://developers.notion.com/reference/property-object#status
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.attributes import Style
from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import (
    DBCellBase,
    DBPropertyBase,
    FromJSONMixin,
)


@dataclass
class StatusOption(FromJSONMixin):
    color: str
    id: str
    name: str
    description: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class StatusGroup(FromJSONMixin):
    color: str
    id: str
    name: str
    option_ids: List[str] = field(default_factory=List[str])

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class StatusProp(FromJSONMixin):
    options: List[StatusOption] = field(default_factory=list)
    groups: List[StatusGroup] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            options=[StatusOption.from_dict(o) for o in data.get("options", [])],
            groups=[StatusGroup.from_dict(g) for g in data.get("groups", [])],
        )


@dataclass
class Status(DBPropertyBase):
    id: str
    name: str
    status: StatusProp
    type: str = "status"
    description: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(status=StatusProp.from_dict(data.pop("status", {})), **data)


@dataclass
class StatusCell(DBCellBase):
    id: str
    status: Optional[StatusOption]
    type: str = "status"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(status=StatusOption.from_dict(data.pop("status", {})), **data)

    def get_html(self) -> Optional[HtmlTag]:
        if status := self.status:
            select_attr = []
            if status.color and status.color != "default":
                select_attr.append(Style(f"color: {status.color}"))
            return Div(select_attr, status.name)
        return None



from __future__ import annotations

from unstructured_ingest.processes.connector_registry import (
    add_destination_entry,
)

from .cloud import CONNECTOR_TYPE as CLOUD_CONNECTOR_TYPE
from .cloud import qdrant_cloud_destination_entry
from .local import CONNECTOR_TYPE as LOCAL_CONNECTOR_TYPE
from .local import qdrant_local_destination_entry
from .server import CONNECTOR_TYPE as SERVER_CONNECTOR_TYPE
from .server import qdrant_server_destination_entry

add_destination_entry(destination_type=CLOUD_CONNECTOR_TYPE, entry=qdrant_cloud_destination_entry)
add_destination_entry(destination_type=SERVER_CONNECTOR_TYPE, entry=qdrant_server_destination_entry)
add_destination_entry(destination_type=LOCAL_CONNECTOR_TYPE, entry=qdrant_local_destination_entry)



from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generator

from pydantic import Field, Secret

from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.weaviate.weaviate import (
    WeaviateAccessConfig,
    WeaviateConnectionConfig,
    WeaviateUploader,
    WeaviateUploaderConfig,
    WeaviateUploadStager,
    WeaviateUploadStagerConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from weaviate.client import WeaviateClient

CONNECTOR_TYPE = "weaviate-local"


class LocalWeaviateAccessConfig(WeaviateAccessConfig):
    pass


class LocalWeaviateConnectionConfig(WeaviateConnectionConfig):
    access_config: Secret[WeaviateAccessConfig] = Field(
        default=WeaviateAccessConfig(), validate_default=True
    )

    @contextmanager
    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_client(self) -> Generator["WeaviateClient", None, None]:
        from weaviate import connect_to_local
        from weaviate.classes.init import AdditionalConfig

        with connect_to_local(
            additional_config=AdditionalConfig(timeout=self.get_timeout())
        ) as weaviate_client:
            yield weaviate_client


class LocalWeaviateUploadStagerConfig(WeaviateUploadStagerConfig):
    pass


@dataclass
class LocalWeaviateUploadStager(WeaviateUploadStager):
    upload_stager_config: LocalWeaviateUploadStagerConfig = field(
        default_factory=lambda: WeaviateUploadStagerConfig()
    )


class LocalWeaviateUploaderConfig(WeaviateUploaderConfig):
    pass


@dataclass
class LocalWeaviateUploader(WeaviateUploader):
    upload_config: LocalWeaviateUploaderConfig
    connector_type: str = CONNECTOR_TYPE
    connection_config: LocalWeaviateConnectionConfig


weaviate_local_destination_entry = DestinationRegistryEntry(
    connection_config=LocalWeaviateConnectionConfig,
    uploader=LocalWeaviateUploader,
    uploader_config=LocalWeaviateUploaderConfig,
    upload_stager=LocalWeaviateUploadStager,
    upload_stager_config=LocalWeaviateUploadStagerConfig,
)



import asyncio
import hashlib
import json
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, TypedDict, TypeVar

from unstructured_ingest.data_types.file_data import FileData, file_data_from_file
from unstructured_ingest.interfaces import Downloader, download_responses
from unstructured_ingest.logger import logger
from unstructured_ingest.pipeline.interfaces import PipelineStep
from unstructured_ingest.utils.pydantic_models import serialize_base_model_json

DownloaderT = TypeVar("DownloaderT", bound=Downloader)

STEP_ID = "download"


class DownloadStepResponse(TypedDict):
    file_data_path: str
    path: str


@dataclass
class DownloadStep(PipelineStep):
    process: DownloaderT
    identifier: str = STEP_ID

    def __str__(self):
        return f"{self.identifier} ({self.process.__class__.__name__})"

    def __post_init__(self):
        config = (
            self.process.download_config.model_dump_json() if self.process.download_config else None
        )
        connection_config = (
            self.process.connection_config.model_dump_json()
            if self.process.connection_config
            else None
        )
        logger.info(
            f"Created {self.identifier} with configs: {config}, "
            f"connection configs: {connection_config}"
        )

    @staticmethod
    def is_float(value: str):
        try:
            float(value)
            return True
        except ValueError:
            return False

    def should_download(self, file_data: FileData, file_data_path: str) -> bool:
        if self.context.re_download:
            return True
        download_path = self.process.get_download_path(file_data=file_data)
        if not download_path or not download_path.exists():
            return True
        if (
            download_path.is_file()
            and file_data.metadata.date_modified
            and self.is_float(file_data.metadata.date_modified)
            and download_path.stat().st_mtime > float(file_data.metadata.date_modified)
        ):
            # Also update file data to mark this to reprocess since this won't change the filename
            file_data.reprocess = True
            file_data.to_file(path=file_data_path)
            return True
        return False

    def update_file_data(
        self, file_data: FileData, file_data_path: Path, download_path: Path
    ) -> None:
        file_data.local_download_path = str(download_path.resolve())
        file_size_bytes = download_path.stat().st_size
        if not file_data.metadata.filesize_bytes and file_size_bytes:
            file_data.metadata.filesize_bytes = file_size_bytes
        if (
            file_data.metadata.filesize_bytes
            and file_data.metadata.filesize_bytes != file_size_bytes
        ):
            logger.warning(
                f"file size in original file data "
                f"({file_data.metadata.filesize_bytes}) doesn't "
                f"match size of local file: {file_size_bytes}, updating"
            )
            file_data.metadata.filesize_bytes = file_size_bytes
        logger.debug(f"updating file data with new content: {file_data.model_dump_json()}")
        with file_data_path.open("w") as file:
            file.write(file_data.model_dump_json(indent=2))

    async def _run_async(self, fn: Callable, file_data_path: str) -> list[DownloadStepResponse]:
        file_data = file_data_from_file(path=file_data_path)
        download_path = self.process.get_download_path(file_data=file_data)
        if not self.should_download(file_data=file_data, file_data_path=file_data_path):
            logger.debug(f"skipping download, file already exists locally: {download_path}")
            self.update_file_data(
                file_data=file_data,
                file_data_path=Path(file_data_path),
                download_path=download_path,
            )
            return [DownloadStepResponse(file_data_path=file_data_path, path=str(download_path))]
        fn_kwargs = {"file_data": file_data}
        if not asyncio.iscoroutinefunction(fn):
            download_results = fn(**fn_kwargs)
        elif semaphore := self.context.semaphore:
            async with semaphore:
                download_results = await fn(**fn_kwargs)
        else:
            download_results = await fn(**fn_kwargs)
        return self.create_step_results(
            current_file_data_path=file_data_path,
            download_results=download_results,
            current_file_data=file_data,
        )

    def create_step_results(
        self,
        current_file_data_path: str,
        current_file_data: FileData,
        download_results: download_responses,
    ) -> list[DownloadStepResponse]:
        responses = []
        if not isinstance(download_results, list):
            file_data = current_file_data
            file_data_path = current_file_data_path
            download_path = download_results["path"]
            if download_results["file_data"].identifier == current_file_data.identifier:
                self.update_file_data(
                    file_data=file_data,
                    file_data_path=Path(file_data_path),
                    download_path=download_path,
                )
                responses = [
                    DownloadStepResponse(file_data_path=file_data_path, path=str(download_path))
                ]
            else:
                file_data = download_results["file_data"]
                file_data_path = self.persist_new_file_data(file_data=file_data)
                self.update_file_data(
                    file_data=file_data,
                    file_data_path=Path(file_data_path),
                    download_path=download_path,
                )
                responses = [
                    DownloadStepResponse(
                        file_data_path=current_file_data_path, path=str(download_results["path"])
                    )
                ]
        else:
            # Supplemental results generated as part of the download process
            for res in download_results:
                file_data = res["file_data"]
                file_data_path = self.persist_new_file_data(file_data=file_data)
                download_path = res["path"]
                self.update_file_data(
                    file_data=file_data,
                    file_data_path=Path(file_data_path),
                    download_path=download_path,
                )
                responses.append(
                    DownloadStepResponse(file_data_path=file_data_path, path=res["path"])
                )

        return responses

    def persist_new_file_data(self, file_data: FileData) -> str:
        record_hash = self.get_hash(extras=[file_data.identifier])
        filename = f"{record_hash}.json"
        filepath = (self.cache_dir / filename).resolve()
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(str(filepath), "w") as f:
            f.write(file_data.model_dump_json(indent=2))
        return str(filepath)

    def get_hash(self, extras: Optional[list[str]]) -> str:
        download_config_dict = json.loads(
            serialize_base_model_json(model=self.process.download_config)
        )
        connection_config_dict = json.loads(
            serialize_base_model_json(model=self.process.connection_config)
        )
        hashable_dict = {
            "download_config": download_config_dict,
            "connection_config": connection_config_dict,
        }
        hashable_string = json.dumps(hashable_dict, sort_keys=True)
        if extras:
            hashable_string += "".join(extras)
        return hashlib.sha256(hashable_string.encode()).hexdigest()[:12]

    @property
    def cache_dir(self) -> Path:
        return self.process.download_config.download_dir

    def delete_cache(self):
        if (
            self.context.iter_delete
            and not self.context.preserve_downloads
            and self.cache_dir.exists()
        ):
            cache_dir = self.cache_dir
            logger.info(f"deleting {self.identifier} cache dir {cache_dir}")
            shutil.rmtree(cache_dir)



from __future__ import annotations

from unstructured_ingest.processes.connector_registry import add_destination_entry

from .ibm_watsonx_s3 import CONNECTOR_TYPE as IBM_WATSONX_S3_CONNECTOR_TYPE
from .ibm_watsonx_s3 import ibm_watsonx_s3_destination_entry

add_destination_entry(
    destination_type=IBM_WATSONX_S3_CONNECTOR_TYPE, entry=ibm_watsonx_s3_destination_entry
)



# https://developers.notion.com/reference/block#table
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import HtmlTag, Td, Th, Tr

from unstructured_ingest.processes.connectors.notion.interfaces import BlockBase, FromJSONMixin
from unstructured_ingest.processes.connectors.notion.types.rich_text import RichText


@dataclass
class Table(BlockBase):
    table_width: int
    has_column_header: bool
    has_row_header: bool

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return None


@dataclass
class TableCell(FromJSONMixin):
    rich_texts: List[RichText]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(rich_texts=[RichText.from_dict(rt) for rt in data.pop("rich_texts", [])])

    def get_html(self, is_header: bool) -> Optional[HtmlTag]:
        if is_header:
            return Th([], [rt.get_html() for rt in self.rich_texts])
        else:
            return Td([], [rt.get_html() for rt in self.rich_texts])


# https://developers.notion.com/reference/block#table-rows
@dataclass
class TableRow(BlockBase):
    is_header: bool = False
    cells: List[TableCell] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict):
        cells = data.get("cells", [])
        return cls(cells=[TableCell.from_dict({"rich_texts": c}) for c in cells])

    @staticmethod
    def can_have_children() -> bool:
        return False

    def get_html(self) -> Optional[HtmlTag]:
        return Tr([], [cell.get_html(is_header=self.is_header) for cell in self.cells])



# https://developers.notion.com/reference/property-object#checkbox
from dataclasses import dataclass, field
from typing import Optional

from htmlBuilder.attributes import Checked, Type
from htmlBuilder.tags import Div, HtmlTag, Input

from unstructured_ingest.processes.connectors.notion.interfaces import DBCellBase, DBPropertyBase


@dataclass
class Checkbox(DBPropertyBase):
    id: str
    name: str
    type: str = "checkbox"
    description: Optional[str] = None
    checkbox: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class CheckboxCell(DBCellBase):
    id: str
    checkbox: bool
    name: Optional[str] = None
    type: str = "checkbox"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        check_input_attributes = [Type("checkbox")]
        if self.checkbox:
            check_input_attributes.append(Checked(""))
        return Div([], Input(check_input_attributes))



from typing import Dict

from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connectors.notion.interfaces import DBCellBase, DBPropertyBase

from .checkbox import Checkbox, CheckboxCell
from .created_by import CreatedBy, CreatedByCell
from .created_time import CreatedTime, CreatedTimeCell
from .date import Date, DateCell
from .email import Email, EmailCell
from .files import Files, FilesCell
from .formula import Formula, FormulaCell
from .last_edited_by import LastEditedBy, LastEditedByCell
from .last_edited_time import LastEditedTime, LastEditedTimeCell
from .multiselect import MultiSelect, MultiSelectCell
from .number import Number, NumberCell
from .people import People, PeopleCell
from .phone_number import PhoneNumber, PhoneNumberCell
from .relation import Relation, RelationCell
from .rich_text import RichText, RichTextCell
from .rollup import Rollup, RollupCell
from .select import Select, SelectCell
from .status import Status, StatusCell
from .title import Title, TitleCell
from .unique_id import UniqueID, UniqueIDCell
from .url import URL, URLCell
from .verification import Verification, VerificationCell

# It's possible to add 'button' property to Notion database.
# However, current Notion API documentation doesn't mention it.
# Buttons are only functional inside Notion UI. We can simply
# ignore them so that the we don't throw an error when trying
# to map 'button' properties.
unsupported_db_prop_types = ["button"]

db_prop_type_mapping = {
    "checkbox": Checkbox,
    "created_by": CreatedBy,
    "created_time": CreatedTime,
    "date": Date,
    "email": Email,
    "files": Files,
    "formula": Formula,
    "last_edited_by": LastEditedBy,
    "last_edited_time": LastEditedTime,
    "multi_select": MultiSelect,
    "number": Number,
    "people": People,
    "phone_number": PhoneNumber,
    "relation": Relation,
    "rich_text": RichText,
    "rollup": Rollup,
    "select": Select,
    "status": Status,
    "title": Title,
    "unique_id": UniqueID,
    "url": URL,
    "verification": Verification,
}


def map_properties(props: Dict[str, dict]) -> Dict[str, DBPropertyBase]:
    mapped_dict = {}
    for k, v in props.items():
        try:
            property_type = v["type"]
            if property_type in unsupported_db_prop_types:
                logger.warning(
                    f"Unsupported property type '{property_type}' for property '{k}'. Skipping."
                )
                continue
            mapped_dict[k] = db_prop_type_mapping[property_type].from_dict(v)  # type: ignore
        except KeyError as ke:
            raise KeyError(f"failed to map to associated database property -> {k}: {v}") from ke

    return mapped_dict


db_cell_type_mapping = {
    "checkbox": CheckboxCell,
    "created_by": CreatedByCell,
    "created_time": CreatedTimeCell,
    "date": DateCell,
    "email": EmailCell,
    "files": FilesCell,
    "formula": FormulaCell,
    "last_edited_by": LastEditedByCell,
    "last_edited_time": LastEditedTimeCell,
    "multi_select": MultiSelectCell,
    "number": NumberCell,
    "people": PeopleCell,
    "phone_number": PhoneNumberCell,
    "relation": RelationCell,
    "rich_text": RichTextCell,
    "rollup": RollupCell,
    "select": SelectCell,
    "status": StatusCell,
    "title": TitleCell,
    "unique_id": UniqueIDCell,
    "url": URLCell,
    "verification": VerificationCell,
}


def map_cells(props: Dict[str, dict]) -> Dict[str, DBCellBase]:
    mapped_dict = {}
    for k, v in props.items():
        try:
            property_type = v["type"]
            if property_type in unsupported_db_prop_types:
                logger.warning(
                    f"Unsupported property type '{property_type}' for property '{k}'. Skipping."
                )
                continue
            mapped_dict[k] = db_cell_type_mapping[property_type].from_dict(v)  # type: ignore
        except KeyError as ke:
            raise KeyError(f"failed to map to associated database property -> {k}: {v}") from ke

    return mapped_dict


__all__ = [
    "map_properties",
    "map_cells",
]



from dataclasses import dataclass

from pydantic import Field, Secret

from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.qdrant.qdrant import (
    QdrantAccessConfig,
    QdrantConnectionConfig,
    QdrantUploader,
    QdrantUploaderConfig,
    QdrantUploadStager,
    QdrantUploadStagerConfig,
)

CONNECTOR_TYPE = "qdrant-server"


class ServerQdrantAccessConfig(QdrantAccessConfig):
    pass


class ServerQdrantConnectionConfig(QdrantConnectionConfig):
    url: str = Field(default=None, description="url of Qdrant server")
    access_config: Secret[ServerQdrantAccessConfig] = Field(
        default_factory=ServerQdrantAccessConfig, validate_default=True
    )

    def get_client_kwargs(self) -> dict:
        return {
            "url": self.url,
        }


class ServerQdrantUploadStagerConfig(QdrantUploadStagerConfig):
    pass


@dataclass
class ServerQdrantUploadStager(QdrantUploadStager):
    upload_stager_config: ServerQdrantUploadStagerConfig


class ServerQdrantUploaderConfig(QdrantUploaderConfig):
    pass


@dataclass
class ServerQdrantUploader(QdrantUploader):
    connection_config: ServerQdrantConnectionConfig
    upload_config: ServerQdrantUploaderConfig
    connector_type: str = CONNECTOR_TYPE


qdrant_server_destination_entry = DestinationRegistryEntry(
    connection_config=ServerQdrantConnectionConfig,
    uploader=ServerQdrantUploader,
    uploader_config=ServerQdrantUploaderConfig,
    upload_stager=ServerQdrantUploadStager,
    upload_stager_config=ServerQdrantUploadStagerConfig,
)



# https://developers.notion.com/reference/block#to-do
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.attributes import Checked, Style, Type
from htmlBuilder.tags import Div, HtmlTag, Input

from unstructured_ingest.processes.connectors.notion.interfaces import BlockBase
from unstructured_ingest.processes.connectors.notion.types.rich_text import RichText


@dataclass
class ToDo(BlockBase):
    color: str
    checked: bool = False
    rich_text: List[RichText] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        rich_text = data.pop("rich_text", [])
        todo = cls(**data)
        todo.rich_text = [RichText.from_dict(rt) for rt in rich_text]
        return todo

    def get_html(self) -> Optional[HtmlTag]:
        if not self.rich_text:
            return None

        elements = []
        check_input_attributes = [Type("checkbox")]
        if self.checked:
            check_input_attributes.append(Checked(""))
        elements.append(Input(check_input_attributes))
        elements.extend([rt.get_html() for rt in self.rich_text])
        attributes = []
        if self.color and self.color != "default":
            attributes.append(Style(f"color: {self.color}"))
        return Div(attributes, elements)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

from pydantic import Field

from unstructured_ingest.embed.interfaces import (
    EMBEDDINGS_KEY,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from sentence_transformers import SentenceTransformer


class HuggingFaceEmbeddingConfig(EmbeddingConfig):
    embedder_model_name: Optional[str] = Field(
        default="all-MiniLM-L6-v2", alias="model_name", description="HuggingFace model name"
    )
    embedder_model_kwargs: Optional[dict] = Field(
        default_factory=lambda: {"device": "cpu"},
        alias="model_kwargs",
        description="additional model parameters",
    )
    encode_kwargs: Optional[dict] = Field(
        default_factory=lambda: {"normalize_embeddings": False},
        description="additional embedding parameters",
    )

    @requires_dependencies(
        ["sentence_transformers"],
        extras="huggingface",
    )
    def get_client(self) -> "SentenceTransformer":
        from sentence_transformers import SentenceTransformer

        return SentenceTransformer(
            model_name_or_path=self.embedder_model_name,
            **self.embedder_model_kwargs,
        )

    def get_encoder_kwargs(self) -> dict:
        encoder_kwargs = self.encode_kwargs or {}
        encoder_kwargs["batch_size"] = self.batch_size
        return encoder_kwargs


@dataclass
class HuggingFaceEmbeddingEncoder(BaseEmbeddingEncoder):
    config: HuggingFaceEmbeddingConfig = field(default_factory=HuggingFaceEmbeddingConfig)

    def _embed_query(self, query: str) -> list[float]:
        return self._embed_documents(texts=[query])[0]

    def _embed_documents(self, texts: list[str]) -> list[list[float]]:
        client = self.config.get_client()
        embeddings = client.encode(texts, **self.config.get_encoder_kwargs())
        return embeddings.tolist()

    def embed_documents(self, elements: list[dict]) -> list[dict]:
        elements = elements.copy()
        elements_with_text = [e for e in elements if e.get("text")]
        if not elements_with_text:
            return elements
        embeddings = self._embed_documents([e["text"] for e in elements_with_text])
        for element, embedding in zip(elements_with_text, embeddings):
            element[EMBEDDINGS_KEY] = embedding
        return elements



# coding:utf-8
import logging
from collections.abc import Iterable as IterableType
from typing import Any, Iterable, Optional, Type, Union

from backoff import _sync
from backoff._common import _config_handlers, _prepare_logger
from backoff._jitter import full_jitter
from backoff._typing import (
    _Handler,
    _Jitterer,
    _MaybeCallable,
    _MaybeLogger,
    _MaybeSequence,
    _Predicate,
    _WaitGenerator,
)

from unstructured_ingest.processes.connectors.notion.ingest_backoff._common import (
    _log_backoff,
    _log_giveup,
    _log_start,
)


class RetryHandler:
    def __init__(
        self,
        wait_gen: _WaitGenerator,
        exception: _MaybeSequence[Type[Exception]],
        *,
        max_tries: Optional[_MaybeCallable[int]] = None,
        max_time: Optional[_MaybeCallable[float]] = None,
        jitter: Union[_Jitterer, None] = full_jitter,
        giveup: _Predicate[Exception] = lambda e: False,
        on_start: Union[_Handler, Iterable[_Handler], None] = None,
        on_success: Union[_Handler, Iterable[_Handler], None] = None,
        on_backoff: Union[_Handler, Iterable[_Handler], None] = None,
        on_giveup: Union[_Handler, Iterable[_Handler], None] = None,
        raise_on_giveup: bool = True,
        logger: _MaybeLogger = "backoff",
        start_log_level: int = logging.INFO,
        backoff_log_level: int = logging.INFO,
        giveup_log_level: int = logging.ERROR,
        **wait_gen_kwargs: Any,
    ):
        prepared_logger = _prepare_logger(logger)
        on_success = _config_handlers(on_success)
        on_start = _config_handlers(
            on_start,
            default_handler=_log_start,
            logger=prepared_logger,
            log_level=start_log_level,
        )
        on_backoff = _config_handlers(
            on_backoff,
            default_handler=_log_backoff,
            logger=prepared_logger,
            log_level=backoff_log_level,
        )
        on_giveup = _config_handlers(
            on_giveup,
            default_handler=_log_giveup,
            logger=prepared_logger,
            log_level=giveup_log_level,
        )
        prepared_logger.debug(
            "Initiating retry handler with "
            "max_tries={}, "
            "max_time={}, "
            "exception={}, "
            "start_log_level={}, "
            "backoff_log_level={}, "
            "giveup_log_level={}".format(
                max_tries,
                max_time,
                (
                    ", ".join([e.__name__ for e in exception])
                    if isinstance(exception, IterableType)
                    else exception.__name__
                ),
                logging.getLevelName(start_log_level),
                logging.getLevelName(backoff_log_level),
                logging.getLevelName(giveup_log_level),
            ),
        )
        self.on_start = on_start
        self.on_success = on_success
        self.on_backoff = on_backoff
        self.on_giveup = on_giveup
        self.jitter = jitter
        self.giveup = giveup
        self.raise_on_giveup = raise_on_giveup
        self.wait_gen_kwargs = wait_gen_kwargs
        self.wait_gen = wait_gen
        self.exception = exception
        self.max_tries = max_tries
        self.max_time = max_time

    def __call__(self, target, *args, **kwargs):
        _sync._call_handlers(
            self.on_start,
            target=target,
            args=args,
            kwargs=kwargs,
            tries=None,
            elapsed=None,
            max_tries=self.max_tries,
            max_time=self.max_time,
            exception=self.exception,
        )
        wrapped_func = _sync.retry_exception(
            target,
            self.wait_gen,
            self.exception,
            max_tries=self.max_tries,
            max_time=self.max_time,
            jitter=self.jitter,
            giveup=self.giveup,
            on_success=self.on_success,
            on_backoff=self.on_backoff,
            on_giveup=self.on_giveup,
            raise_on_giveup=self.raise_on_giveup,
            wait_gen_kwargs=self.wait_gen_kwargs,
        )
        return wrapped_func(*args, **kwargs)



import asyncio
import json
import os
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncIterable

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    EMBEDDINGS_KEY,
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.errors_v2 import (
    ProviderError,
    RateLimitError,
    UserAuthError,
    UserError,
    is_internal_error,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from botocore.client import BaseClient

    class BedrockRuntimeClient(BaseClient):
        def invoke_model(self, body: str, modelId: str, accept: str, contentType: str) -> dict:
            pass

    class AsyncBedrockRuntimeClient(BaseClient):
        async def invoke_model(
            self, body: str, modelId: str, accept: str, contentType: str
        ) -> dict:
            pass

    class BedrockClient(BaseClient):
        def list_foundation_models(self, byOutputModality: str) -> dict:
            pass


def conform_query(query: str, provider: str) -> dict:
    # replace newlines, which can negatively affect performance.
    text = query.replace(os.linesep, " ")

    # format input body for provider
    input_body = {}
    if provider == "cohere":
        if "input_type" not in input_body:
            input_body["input_type"] = "search_document"
        input_body["texts"] = [text]
    else:
        # includes common provider == "amazon"
        input_body["inputText"] = text
    return input_body


class BedrockEmbeddingConfig(EmbeddingConfig):
    aws_access_key_id: SecretStr = Field(description="aws access key id")
    aws_secret_access_key: SecretStr = Field(description="aws secret access key")
    region_name: str = Field(description="aws region name", default="us-west-2")
    embedder_model_name: str = Field(
        default="amazon.titan-embed-text-v1",
        alias="model_name",
        description="AWS Bedrock model name",
    )

    def wrap_error(self, e: Exception) -> Exception:
        if is_internal_error(e=e):
            return e
        from botocore.exceptions import ClientError

        if isinstance(e, ClientError):
            # https://docs.aws.amazon.com/awssupport/latest/APIReference/CommonErrors.html
            http_response = e.response
            meta = http_response["ResponseMetadata"]
            http_response_code = meta["HTTPStatusCode"]
            error_code = http_response["Error"]["Code"]
            if http_response_code == 400:
                if error_code == "ValidationError":
                    return UserError(http_response["Error"])
                elif error_code == "ThrottlingException":
                    return RateLimitError(http_response["Error"])
                elif error_code == "NotAuthorized" or error_code == "AccessDeniedException":
                    return UserAuthError(http_response["Error"])
            if http_response_code == 403:
                return UserAuthError(http_response["Error"])
            if 400 <= http_response_code < 500:
                return UserError(http_response["Error"])
            if http_response_code >= 500:
                return ProviderError(http_response["Error"])

        logger.error(f"unhandled exception from bedrock: {e}", exc_info=True)
        return e

    def run_precheck(self) -> None:
        client = self.get_bedrock_client()
        try:
            model_info = client.list_foundation_models(byOutputModality="EMBEDDING")
            summaries = model_info.get("modelSummaries", [])
            model_ids = [m["modelId"] for m in summaries]
            arns = [":".join(m["modelArn"]) for m in summaries]

            if self.embedder_model_name not in model_ids and self.embedder_model_name not in arns:
                raise UserError(
                    "model '{}' not found either : {} or {}".format(
                        self.embedder_model_name, ", ".join(model_ids), ", ".join(arns)
                    )
                )
        except Exception as e:
            raise self.wrap_error(e=e)

    def get_client_kwargs(self) -> dict:
        return {
            "aws_access_key_id": self.aws_access_key_id.get_secret_value(),
            "aws_secret_access_key": self.aws_secret_access_key.get_secret_value(),
            "region_name": self.region_name,
        }

    @requires_dependencies(
        ["boto3"],
        extras="bedrock",
    )
    def get_bedrock_client(self) -> "BedrockClient":
        import boto3

        bedrock_client = boto3.client(service_name="bedrock", **self.get_client_kwargs())

        return bedrock_client

    @requires_dependencies(
        ["boto3", "numpy", "botocore"],
        extras="bedrock",
    )
    def get_client(self) -> "BedrockRuntimeClient":
        import boto3

        bedrock_client = boto3.client(service_name="bedrock-runtime", **self.get_client_kwargs())

        return bedrock_client

    @requires_dependencies(
        ["aioboto3"],
        extras="bedrock",
    )
    @asynccontextmanager
    async def get_async_client(self) -> AsyncIterable["AsyncBedrockRuntimeClient"]:
        import aioboto3

        session = aioboto3.Session()
        async with session.client("bedrock-runtime", **self.get_client_kwargs()) as aws_bedrock:
            yield aws_bedrock


@dataclass
class BedrockEmbeddingEncoder(BaseEmbeddingEncoder):
    config: BedrockEmbeddingConfig

    def precheck(self):
        self.config.run_precheck()

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def embed_query(self, query: str) -> list[float]:
        """Call out to Bedrock embedding endpoint."""
        provider = self.config.embedder_model_name.split(".")[0]
        body = conform_query(query=query, provider=provider)

        bedrock_client = self.config.get_client()
        # invoke bedrock API
        try:
            response = bedrock_client.invoke_model(
                body=json.dumps(body),
                modelId=self.config.embedder_model_name,
                accept="application/json",
                contentType="application/json",
            )
        except Exception as e:
            raise self.wrap_error(e=e)

        # format output based on provider
        response_body = json.loads(response.get("body").read())
        if provider == "cohere":
            return response_body.get("embeddings")[0]
        else:
            # includes common provider == "amazon"
            return response_body.get("embedding")

    def embed_documents(self, elements: list[dict]) -> list[dict]:
        elements = elements.copy()
        elements_with_text = [e for e in elements if e.get("text")]
        if not elements_with_text:
            return elements
        embeddings = [self.embed_query(query=e["text"]) for e in elements_with_text]
        for element, embedding in zip(elements_with_text, embeddings):
            element[EMBEDDINGS_KEY] = embedding
        return elements


@dataclass
class AsyncBedrockEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: BedrockEmbeddingConfig

    def precheck(self):
        self.config.run_precheck()

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    async def embed_query(self, query: str) -> list[float]:
        """Call out to Bedrock embedding endpoint."""
        provider = self.config.embedder_model_name.split(".")[0]
        body = conform_query(query=query, provider=provider)
        try:
            async with self.config.get_async_client() as bedrock_client:
                # invoke bedrock API
                try:
                    response = await bedrock_client.invoke_model(
                        body=json.dumps(body),
                        modelId=self.config.embedder_model_name,
                        accept="application/json",
                        contentType="application/json",
                    )
                except Exception as e:
                    raise self.wrap_error(e=e)
                async with response.get("body") as client_response:
                    response_body = await client_response.json()

            # format output based on provider
            if provider == "cohere":
                return response_body.get("embeddings")[0]
            else:
                # includes common provider == "amazon"
                return response_body.get("embedding")
        except Exception as e:
            raise ValueError(f"Error raised by inference endpoint: {e}")

    async def embed_documents(self, elements: list[dict]) -> list[dict]:
        elements = elements.copy()
        elements_with_text = [e for e in elements if e.get("text")]
        embeddings = await asyncio.gather(
            *[self.embed_query(query=e.get("text", "")) for e in elements_with_text]
        )
        for element, embedding in zip(elements_with_text, embeddings):
            element[EMBEDDINGS_KEY] = embedding
        return elements



from dataclasses import dataclass

from pydantic import Field, Secret

from unstructured_ingest.interfaces.connector import AccessConfig
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.lancedb.lancedb import (
    LanceDBRemoteConnectionConfig,
    LanceDBUploader,
    LanceDBUploaderConfig,
    LanceDBUploadStager,
    LanceDBUploadStagerConfig,
)

CONNECTOR_TYPE = "lancedb_cloud"


class LanceDBCloudAccessConfig(AccessConfig):
    api_key: str = Field(description="Api key associated with LanceDb cloud")


class LanceDBCloudConnectionConfig(LanceDBRemoteConnectionConfig):
    access_config: Secret[LanceDBCloudAccessConfig]

    def get_storage_options(self) -> dict:
        return {**self.access_config.get_secret_value().model_dump(), "timeout": self.timeout}


@dataclass
class LanceDBCloudUploader(LanceDBUploader):
    upload_config: LanceDBUploaderConfig
    connection_config: LanceDBCloudConnectionConfig
    connector_type: str = CONNECTOR_TYPE


lancedb_cloud_destination_entry = DestinationRegistryEntry(
    connection_config=LanceDBCloudConnectionConfig,
    uploader=LanceDBCloudUploader,
    uploader_config=LanceDBUploaderConfig,
    upload_stager_config=LanceDBUploadStagerConfig,
    upload_stager=LanceDBUploadStager,
)



import socket
from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import Field, Secret, SecretStr

from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.processes.connectors.kafka.kafka import (
    KafkaAccessConfig,
    KafkaConnectionConfig,
    KafkaDownloader,
    KafkaDownloaderConfig,
    KafkaIndexer,
    KafkaIndexerConfig,
    KafkaUploader,
    KafkaUploaderConfig,
)

if TYPE_CHECKING:
    pass

CONNECTOR_TYPE = "kafka-cloud"


class CloudKafkaAccessConfig(KafkaAccessConfig):
    kafka_api_key: SecretStr = Field(
        description="Kafka API key to connect at the server", default=None
    )
    secret: SecretStr = Field(description="", default=None)


class CloudKafkaConnectionConfig(KafkaConnectionConfig):
    access_config: Secret[CloudKafkaAccessConfig]

    def get_consumer_configuration(self) -> dict:
        bootstrap = self.bootstrap_server
        port = self.port
        access_config = self.access_config.get_secret_value()

        conf = {
            "bootstrap.servers": f"{bootstrap}:{port}",
            "client.id": socket.gethostname(),
            "group.id": self.group_id,
            "enable.auto.commit": "false",
            "auto.offset.reset": "earliest",
            "sasl.username": access_config.kafka_api_key.get_secret_value(),
            "sasl.password": access_config.secret.get_secret_value(),
            "sasl.mechanism": "PLAIN",
            "security.protocol": "SASL_SSL",
            "logger": logger,
        }

        return conf

    def get_producer_configuration(self) -> dict:
        bootstrap = self.bootstrap_server
        port = self.port
        access_config = self.access_config.get_secret_value()

        conf = {
            "bootstrap.servers": f"{bootstrap}:{port}",
            "sasl.username": access_config.kafka_api_key.get_secret_value(),
            "sasl.password": access_config.secret.get_secret_value(),
            "sasl.mechanism": "PLAIN",
            "security.protocol": "SASL_SSL",
            "logger": logger,
        }

        return conf


class CloudKafkaIndexerConfig(KafkaIndexerConfig):
    pass


@dataclass
class CloudKafkaIndexer(KafkaIndexer):
    connection_config: CloudKafkaConnectionConfig
    index_config: CloudKafkaIndexerConfig
    connector_type: str = CONNECTOR_TYPE


class CloudKafkaDownloaderConfig(KafkaDownloaderConfig):
    pass


@dataclass
class CloudKafkaDownloader(KafkaDownloader):
    connection_config: CloudKafkaConnectionConfig
    download_config: CloudKafkaDownloaderConfig
    connector_type: str = CONNECTOR_TYPE


class CloudKafkaUploaderConfig(KafkaUploaderConfig):
    pass


@dataclass
class CloudKafkaUploader(KafkaUploader):
    connection_config: CloudKafkaConnectionConfig
    upload_config: CloudKafkaUploaderConfig
    connector_type: str = CONNECTOR_TYPE


kafka_cloud_source_entry = SourceRegistryEntry(
    connection_config=CloudKafkaConnectionConfig,
    indexer=CloudKafkaIndexer,
    indexer_config=CloudKafkaIndexerConfig,
    downloader=CloudKafkaDownloader,
    downloader_config=CloudKafkaDownloaderConfig,
)

kafka_cloud_destination_entry = DestinationRegistryEntry(
    connection_config=CloudKafkaConnectionConfig,
    uploader=CloudKafkaUploader,
    uploader_config=CloudKafkaUploaderConfig,
)



from abc import ABC
from dataclasses import dataclass
from typing import Any, TypeVar, Union

from pydantic import BaseModel, Secret, model_validator
from pydantic.types import _SecretBase

from unstructured_ingest.processes.utils.logging.connector import ConnectorLoggingMixin


class AccessConfig(BaseModel):
    """Meant to designate holding any sensitive information associated with other configs
    and also for access specific configs."""


AccessConfigT = TypeVar("AccessConfigT", bound=AccessConfig)


class ConnectionConfig(BaseModel):
    access_config: Secret[AccessConfigT]

    def get_access_config(self) -> dict[str, Any]:
        if not self.access_config:
            return {}
        return self.access_config.get_secret_value().model_dump()

    @model_validator(mode="after")
    def check_access_config(self):
        access_config = self.access_config
        if self._is_access_config_optional() and access_config is None:
            return self
        if not isinstance(access_config, _SecretBase):
            raise ValueError("access_config must be an instance of SecretBase")
        return self

    def _is_access_config_optional(self) -> bool:
        access_config_type = self.model_fields["access_config"].annotation
        return (
            hasattr(access_config_type, "__origin__")
            and hasattr(access_config_type, "__args__")
            and access_config_type.__origin__ is Union
            and len(access_config_type.__args__) == 2
            and type(None) in access_config_type.__args__
        )


ConnectionConfigT = TypeVar("ConnectionConfigT", bound=ConnectionConfig)


@dataclass
class BaseConnector(ABC, ConnectorLoggingMixin):
    connection_config: ConnectionConfigT

    def __post_init__(self):
        """Initialize the logging mixin after dataclass initialization."""
        ConnectorLoggingMixin.__init__(self)



from dataclasses import dataclass

from pydantic import Field, Secret

from unstructured_ingest.interfaces.connector import AccessConfig
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.lancedb.lancedb import (
    LanceDBConnectionConfig,
    LanceDBUploader,
    LanceDBUploaderConfig,
    LanceDBUploadStager,
    LanceDBUploadStagerConfig,
)

CONNECTOR_TYPE = "lancedb_local"


class LanceDBLocalAccessConfig(AccessConfig):
    pass


class LanceDBLocalConnectionConfig(LanceDBConnectionConfig):
    access_config: Secret[LanceDBLocalAccessConfig] = Field(
        default_factory=LanceDBLocalAccessConfig, validate_default=True
    )

    def get_storage_options(self) -> None:
        return None


@dataclass
class LanceDBLocalUploader(LanceDBUploader):
    upload_config: LanceDBUploaderConfig
    connection_config: LanceDBLocalConnectionConfig
    connector_type: str = CONNECTOR_TYPE


lancedb_local_destination_entry = DestinationRegistryEntry(
    connection_config=LanceDBLocalConnectionConfig,
    uploader=LanceDBLocalUploader,
    uploader_config=LanceDBUploaderConfig,
    upload_stager_config=LanceDBUploadStagerConfig,
    upload_stager=LanceDBUploadStager,
)



# https://developers.notion.com/reference/block#callout
from dataclasses import dataclass, field
from typing import List, Optional, Union

from htmlBuilder.attributes import Href, Style
from htmlBuilder.tags import A, Div, HtmlTag, P

from unstructured_ingest.processes.connectors.notion.interfaces import (
    BlockBase,
    FromJSONMixin,
    GetHTMLMixin,
)
from unstructured_ingest.processes.connectors.notion.types.rich_text import RichText


@dataclass
class EmojiIcon(FromJSONMixin, GetHTMLMixin):
    emoji: str
    type: str = "emoji"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return P([], self.emoji)


@dataclass
class ExternalIconContent(FromJSONMixin):
    url: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class ExternalIcon(FromJSONMixin, GetHTMLMixin):
    external: ExternalIconContent
    type: str = "external"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(external=ExternalIconContent.from_dict(data=data.pop("external")), **data)

    def get_html(self) -> Optional[HtmlTag]:
        if self.external:
            return A([Href(self.external.url)], [self.external.url])
        else:
            return None


@dataclass
class FileIconContent(FromJSONMixin):
    url: str
    expiry_time: Optional[str] = None  # Add expiry_time if needed

    @classmethod
    def from_dict(cls, data: dict):
        # Only include expiry_time if it exists in the dictionary
        # Notion API might not always include it
        init_data = {"url": data.get("url")}
        if "expiry_time" in data:
            init_data["expiry_time"] = data.get("expiry_time")
        return cls(**init_data)


@dataclass
class FileIcon(FromJSONMixin, GetHTMLMixin):
    file: FileIconContent
    type: str = "file"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(file=FileIconContent.from_dict(data=data.pop("file")), **data)

    def get_html(self) -> Optional[HtmlTag]:
        # Render the file URL, similar to how ExternalIcon is handled
        if self.file:
            # Could potentially render an <img> tag, but sticking to URL for consistency
            return A([Href(self.file.url)], [f"[File Icon: {self.file.url}]"])
        else:
            return None


class Icon(FromJSONMixin):
    @classmethod
    def from_dict(cls, data: dict) -> Union[EmojiIcon, ExternalIcon, FileIcon]:
        t = data.get("type")
        if t == "emoji":
            return EmojiIcon.from_dict(data)
        elif t == "external":
            return ExternalIcon.from_dict(data)
        elif t == "file":
            return FileIcon.from_dict(data)
        else:
            raise ValueError(f"Unexpected icon type: {t} ({data})")


@dataclass
class Callout(BlockBase):
    color: str
    icon: Optional[Union[EmojiIcon, ExternalIcon, FileIcon]] = None
    rich_text: List[RichText] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        rich_text = data.pop("rich_text", [])
        icon_data = data.pop("icon", None)
        icon = Icon.from_dict(icon_data) if icon_data else None
        return cls(
            color=data["color"],
            icon=icon,
            rich_text=[RichText.from_dict(rt) for rt in rich_text],
        )

    def get_html(self) -> Optional[HtmlTag]:
        elements = []
        if self.icon and self.icon.get_html():
            elements.append(self.icon.get_html())
        if self.rich_text:
            elements.extend([rt.get_html() for rt in self.rich_text])
        attributes = []
        if self.color:
            attributes.append(Style(f"color:{self.color}"))
        return Div(attributes, elements)
